from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Resource

@Resource("batch_job")
class DuploBatchSchedulingPolicy(DuploTenantResourceV3):
  """Manage AWS Batch Scheduling Policies

  Run batch jobs as a managed service on AWS infrastructure. 

  Read more docs here: 
  https://docs.duplocloud.com/docs/overview/aws-services/batch
  """

  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, 
                     slug="aws/batchSchedulingPolicy",
                     prefixed=True)

  def name_from_body(self, body):
    return body["Name"]

  


import time
from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.errors import DuploError, DuploFailedResource, DuploStillWaiting
from duplocloud.commander import Command, Resource
from json import dumps, loads
import duplocloud.args as args

_STATUS_CODES = {
  "1": "Running",
  "3": "Pending",
  "4": "Waiting",
  "6": "Deleted",
  "7": "Failed",
  "11": "Succeeded"
}

@Resource("service")
class DuploService(DuploTenantResourceV2):
  """Duplocloud Service Resource
  
  This resource is used to manage services in Duplocloud. Using the `duploctl` command line tool, you can manage services with actions:
  
  Usage: Basic CLI Use  
    ```sh
    duploctl service <action>
    ```
  """
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
    self.paths = {
      "list": "GetReplicationControllers"
    }
    self.__pod_svc = self.duplo.load("pod")
    self.__old = None
    self.__updates = None
    self.__old_replicaset = None

  def _validate_args(self, name, all, targets):
    if sum(bool(x) for x in [name, all, targets]) > 1:
      raise DuploError("You can only specify one of: a service name, the '--all' flag, or the '--targets' flag.")
    if not any([name, all, targets]):
      raise DuploError("You must specify a service name, use the '--all' flag, or provide '--targets'.")

  def _process_service(self, action, service_name, results, wait):
    """Helper function to process a service action (start/stop) and handle errors."""
    try:
      endpoint_action = "ReplicationController" + action.capitalize()
      response = self.duplo.post(self.endpoint(f"{endpoint_action}/{service_name}"))

      if response.status_code == 200:
        results["success"].append(service_name)
        if wait:
          service = self.find(service_name)
          self.wait(service, service)
      else:
        results["errors"].append(
          {"service": service_name, "error": f"Unexpected response: {response}"}
        )
    except Exception as e:
      results["errors"].append({"service": service_name, "error": str(e)})

  def _perform_action(self, action, name=None, all=False, targets=None, wait=False):
    """Generic method to perform a start or stop action on services."""
    self._validate_args(name, all, targets)

    results = {"success": [], "errors": []}

    if all:
      services = self.list()
      for service in services:
        service_name = service["Name"]
        self._process_service(action, service_name, results, wait)
    elif targets:
      for service_name in targets:
        self._process_service(action, service_name, results, wait)
    else:
      self._process_service(action, name, results, wait)

    return {
      "message": f"Service {action} operation completed.",
      "details": results,
    }
  
  @Command()
  def find(self, 
           name: args.NAME) -> dict:
    """Find a service by name.
    
    First we try for the endpoint that gives one by name. Otherwise we default to finding it in the list.
    """
    try:
      endpoint = f"v3/subscriptions/{self.tenant_id}/replicationcontroller/{name}"
      response = self.duplo.get(endpoint)
      self.duplo.logger.debug(f"Found service {name} using new endpoint.")
      return response.json()
    # catch the DuploError and let super take over if it's just a 404 which means the new endpoint doesn't exist
    except DuploError:
      self.duplo.logger.debug(f"Service {name} not found using new endpoint, falling back to list.")
      return super().find(name)
    
  @Command()
  def update(self, 
             name: args.NAME,
             body: args.BODY = None,
             patches: args.PATCHES = None) -> dict:
    """Update a service.
    
    Update the state of a service.

    Usage: Basic CLI Use
      Update the replicas to 3 for a service.
      ```sh
      duploctl service update <service-name> --replace Replicas 3 
      ```
    
    Args:
      name: The name of the service to update.
      body: The body of the service to update.
      patches: A list of JSON patches as args to apply to the service.
        The options are `--add`, `--remove`, `--replace`, `--move`, and `--copy`.
        Then followed by `<path>` and `<value>` for `--add`, `--replace`, and `--test`.
      wait: Whether to wait for the service to update.
    
    """
    if (name is None and body is None):
      raise DuploError("No arguments provided to update service", 400)
    if name and body:
      body["Name"] = name
    if body is None:
      body = self.find(name)
    # before patching, save the original body if we need to wait
    old = None
    if self.duplo.wait:
      old = self.find(name) # should use the cache
      old["Replicaset"] = self.current_replicaset(name)
    if patches:
      body = self.duplo.jsonpatch(body, patches)
    if ((ttags := body["Template"].get("AllocationTags", None))
        and not body.get("AllocationTags", None)):
      body["AllocationTags"] = ttags
    body["OtherDockerConfig"] = body["Template"]["OtherDockerConfig"]
    body["AgentPlatform"] = body["Template"].get("AgentPlatform", 0)
    self.duplo.post(self.endpoint("ReplicationControllerChangeAll"), body)
    if self.duplo.wait:
      self.wait(old, body)
    return {
      "message": f"Successfully updated service '{body['Name']}'"
    }
  
  @Command()
  def create(self,
             body: args.BODY) -> dict:
    """Create a service.
    
    Create a service in Duplocloud.
    
    Usage: Basic CLI Use
      ```sh
      duploctl service create --file service.yaml
      ```
      Contents of the `service.yaml` file
      ```yaml
      --8<-- "src/tests/data/service.yaml"
      ```

    Args:
      body: The service to create.
      wait: Wait for the service to be created.

    Returns:
      message: Success message.
    """
    self.duplo.post(self.endpoint("ReplicationControllerUpdate"), body)
    if self.duplo.wait:
      self.wait(lambda: self.find(body["Name"]))
    return {
      "message": f"Successfully created service '{body['Name']}'"
    }

  @Command()
  def delete(self,
             name: args.NAME) -> dict:
    """Delete a service.
    
    Delete a service in Duplocloud.

    Usage: Basic CLI Use
      ```sh
      duploctl service delete <service-name>
      ```

    Args:
      name: The name of the service to delete.

    Returns:
      message: Success message.
    """
    body = {
      "Name": name,
      "State": "delete"
    }
    self.duplo.post(self.endpoint("ReplicationControllerUpdate"), body)
    return {
      "message": f"Successfully deleted service '{name}'"
    }

  @Command()
  def update_replicas(self, 
                      name: args.NAME,
                      replica: args.REPLICAS) -> dict:
    """Scale Service

    Update the number of replicas for a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service update_replicas <service-name> <replicas>
      ```

    Args:
        name (str): The name of the service to update.
        replica (str): Number of replicas to set for service.
    """
    service = self.find(name)
    data = {
      "Name": name,
      "Replicas": replica,
      "AllocationTags": service["Template"].get("AllocationTags", "")
    }
    self.duplo.post(self.endpoint("ReplicationControllerChange"), data)
    if self.duplo.wait:
      self.wait(service, data)
    return {"message": f"Successfully updated replicas for service '{name}'"} 
  
  @Command()
  def update_image(self, 
                   name: args.NAME, 
                   image: args.IMAGE = None,
                   container_image: args.CONTAINER_IMAGE = None,
                   init_container_image: args.INIT_CONTAINER_IMAGE = None) -> dict:
    """Update the image of a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service update_image <service-name> <service-image>
      duploctl service update_image <service-name> --container-image <side-car-container> <container-image>
      duploctl service update_image <service-name> --init-container-image <init-container> <init-container-image>
      ```

    Args:
      name: The name of the service to update.
      image: The new image to use for the service.
      container_image: A list of key-value pairs to set as sidecar container image.
      init_container_image: A list of key-value pairs to set as init container image.
      wait: Whether to wait for the update to complete.

    Returns:
      message: Success message
    """
    if [image, container_image, init_container_image].count(None) != 2:
      raise DuploError("Provide a service image, container images, or init container images.")

    service = self.find(name)
    data = {}
    updated_containers = []
    not_found_containers = []

    if not image:
      other_docker_config = loads(service["Template"].get("OtherDockerConfig", "{}"))
      if container_image:
        images = container_image
        containers = other_docker_config.get("additionalContainers", [])
      elif init_container_image:
        images = init_container_image
        containers = other_docker_config.get("initContainers", [])

      for key, value in images:
        container_found = False
        for c in containers:
          if c["name"] == key:
            c["image"] = value
            updated_containers.append(key)
            container_found = True
            break
        if not container_found:
          not_found_containers.append(key)

      if not updated_containers:
        raise DuploError(f"No matching containers found in service '{name}'")

      data = {
        "Name": name,
        "OtherDockerConfig": dumps(other_docker_config),
        "AllocationTags": service["Template"].get("AllocationTags", "")
      }

    else:
      data = {
        "Name": name,
        "Image": image,
        "AllocationTags": service["Template"].get("AllocationTags", "")
      }

    self.duplo.post(self.endpoint("ReplicationControllerChange"), data)

    if self.duplo.wait:
      self.wait(service, data)

    response_message = "Successfully updated image for service."
    if updated_containers:
      response_message += f" Updated containers: {', '.join(updated_containers)}."
    if not_found_containers:
      response_message += f" Could not find containers: {', '.join(not_found_containers)}."

    return {"message": response_message}
 
  @Command()
  def update_env(self, 
                 name: args.NAME,
                 setvar: args.SETVAR,
                 strategy: args.STRATEGY,
                 deletevar: args.DELETEVAR) -> dict:
    """Update the environment variables of a service. If service has no environment variables set, use -strat replace to set new values.
    Usage: Basic CLI Use
      ```sh
      duploctl service update_env <service-name> --setvar env-key1 env-val1 --setvar env-key2 env-val2 --setvar env-key3 env-val3 -strat merge --host $DUPLO_HOST --tenant $DUPLO_TENANT --token $DUPLO_TOKEN
      ```
    Args:
      name (str): The name of the service to update.

      setvar/-V (list): A list of key value pairs to set as environment variables.
      
      strategy/strat (str): The merge strategy to use for env vars. Valid options are "merge" or "replace".  Default is merge.

      deletevar/-D (list): A list of keys to delete from the environment variables.
    """
    # Returns an array of key and value mappings with provided keys
    def new_env_vars(setvar, key_name="Name", value_name="Value"):
      return [{key_name: i[0], value_name: i[1]} for i in setvar]
  
    def detect_case_format(env_list):
      # Check if the environment variables are using upper, lower, or mixed case for Name and Value
      if all('Name' in env and 'Value' in env for env in env_list):
          return "title"
      elif all('name' in env and 'value' in env for env in env_list):
          return "lower"
      # No consistent standard
      return "mixed"
    
    service = self.find(name)
    currentDockerconfig = loads(service["Template"]["OtherDockerConfig"])
    currentEnv = currentDockerconfig.get("Env", [])
    # Check if user is attempting to merge against a null Env. If so, set currentEnv to empty.
    if currentEnv is None and strategy == "merge":
      self.duplo.logger.warn("Specified \"merge\" strategy on a service with"
                             " no environment variables defined, should use "
                             " \"replace\". Proceeding anyway")
      currentEnv = []
    case_format = detect_case_format(currentEnv)
    if strategy == 'merge':
      try:
        if case_format == "title":
          newEnv = new_env_vars(setvar) if setvar is not None else []
          d = {env['Name']: env for env in currentEnv + newEnv}
        elif case_format == "lower":
          newEnv = new_env_vars(setvar, key_name="name", value_name="value") if setvar is not None else []
          d = {env['name']: env for env in currentEnv + newEnv}
        else:
          self.duplo.logger.warn("Possible attempt to merge env vars with"
                                 " inconsistent Name/Value key case."
                                 " Normalzing to capitalized"
                                 " \"Name\" and \"Value\"")
          norm_currentEnv = [{k.capitalize(): v for k, v in env.items()} for env in currentEnv]
          newEnv = new_env_vars(setvar,) if setvar is not None else []
          d = {env['Name']: env for env in norm_currentEnv + newEnv}
      except KeyError:
        raise DuploError("Could not merge new and existing environment variables")
      mergedvars = list(d.values())
      currentDockerconfig['Env'] = mergedvars
    else:
      newEnv = new_env_vars(setvar) if setvar is not None else []
      currentDockerconfig['Env'] = newEnv
    if deletevar is not None:
      for key in deletevar:
        try:
          currentDockerconfig['Env'] = [d for d in currentDockerconfig['Env'] if d['Name'] != key]
        except KeyError:
          currentDockerconfig['Env'] = [d for d in currentDockerconfig['Env'] if d['name'] != key] 
    payload = {
      "Name": name,
      "OtherDockerConfig": dumps(currentDockerconfig),
      "allocationTags": service["Template"].get("AllocationTags", "")
    }
    self.duplo.post(self.endpoint("ReplicationControllerChange"), payload)
    if self.duplo.wait:
      service["Replicaset"] = self.current_replicaset(name)
      self.wait(service, payload)
    return {"message": f"Successfully updated environment variables for service '{name}'"}
    
  @Command()
  def update_pod_label(self,
                 name: args.NAME,
                 setvar: args.SETVAR,
                 strategy: args.STRATEGY,
                 deletevar: args.DELETEVAR):
  
    """Update the pod labels of a service. If service has no pod labels set, use -strat replace to set new values.
    Usage: Basic CLI Use
      ```sh
      duploctl service update_pod_label <service-name> --setvar env-key1 env-val1 --setvar env-key2 env-val2 --setvar env-key3 env-val3 -strat merge --host $DUPLO_HOST --tenant $DUPLO_TENANT --token $DUPLO_TOKEN
      duploctl service update_pod_label <service-name> --setvar env-key1 env-val1 --setvar env-key2 env-val2 -strat replace --host $DUPLO_HOST --tenant $DUPLO_TENANT --token $DUPLO_TOKEN
      duploctl service update_pod_label <service-name> --deletevar env-key1 --host $DUPLO_HOST --tenant $DUPLO_TENANT --token $DUPLO_TOKEN
      
      ```
    Args:
      name: The name of the service to update.
      setvar/-V: A list of key value pairs to set as environment variables.
      strategy/strat: The merge strategy to use for env vars. Valid options are "merge" or "replace".  Default is merge.
      deletevar/-D: A list of keys to delete from the environment variables.
    """
    service = self.find(name)
    currentDockerconfig = loads(service["Template"]["OtherDockerConfig"])
    currentLabels = currentDockerconfig.get("PodLabels", [])
    newLabels = []
    if setvar is not None:
      newLabels = [{"Name": i[0], "Value": i[1]} for i in setvar]

    # merge new labels in existing labels (append)
    if strategy == 'merge':
      mergedLabels = currentLabels
      for vars in newLabels:
        mergedLabels[vars["Name"]] = vars["Value"]
      currentDockerconfig['PodLabels'] = mergedLabels

    else:
      # replace values of existing label
      for label, value in currentDockerconfig['PodLabels'].items():
        for vars in newLabels:
          if vars["Name"] not in currentDockerconfig['PodLabels']:
            label_name = vars["Name"]
            raise DuploError(f"Service {name} does not have Pod label '{label_name}', enter correct label.", 400)
          if vars["Name"] == label:
            currentDockerconfig['PodLabels'][label] = vars["Value"]

    if deletevar is not None:
      for key in deletevar:
        del currentDockerconfig['PodLabels'][key]
      
    payload = {
      "Name": name,
      "OtherDockerConfig": dumps(currentDockerconfig),
      "allocationTags": service["Template"].get("AllocationTags", "")
    }
    self.duplo.post(self.endpoint("ReplicationControllerChange"), payload)
    if self.duplo.wait:
      service["Replicaset"] = self.current_replicaset(name)
      self.wait(service, payload)
    return {"message": f"Successfully updated pod labels for service '{name}'"}
  
  @Command()
  def bulk_update_image(self, 
                        serviceimage: args.SERVICEIMAGE) -> dict:
    """
    Bulk update the image of multiple services.

    Usage: Basic CLI Use
      ```sh
      duploctl service bulk_update_image -S <service-name-1> <image-name-1> -S <service-name-2> <image-name-2>
      ```

    Args:
      serviceimage: Takes n sets of two arguments, service name and image name. 
                    e.g., -S service1 image1:tag -S service2 image2:tag
      wait: Boolean flag to wait for service updates.
    """
    payload = []
    wait_list = []
    for name, image in serviceimage:
      service = self.find(name)
      payload_item = {
          "Name": name,
          "Image": image,
          "AllocationTags": service["Template"]["AllocationTags"]
      }
      payload.append(payload_item)
      if self.duplo.wait:
        wait_list.append({
          "old": service,
          "updated": payload_item
        })

    self.duplo.post(self.endpoint("ReplicationControllerBulkChangeAll"), payload)

    if self.duplo.wait:
      for update_info in wait_list:
        self.wait(**update_info)

    return {"message": "Successfully updated images for services"}

  @Command()
  def restart(self, 
              name: args.NAME) -> dict:
    """Restart a service.

    Restart a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service restart <service-name>
      ```
    
    Args:
      name: The name of the service to restart.
      wait: Boolean flag to wait for service updates.

    Returns: 
      A success message if the service was restarted successfully.

    Raises:
      DuploError: If the service could not be restarted.
    """
    self.duplo.post(self.endpoint(f"ReplicationControllerReboot/{name}"))
    if self.duplo.wait:
      service = self.find(name)
      self.wait(service, service)
    return {"message": f"Successfully restarted service '{name}'"}
  
  @Command()
  def stop(self,
           name: args.NAME = None,
           all: args.ALL = False,
           targets: args.TARGETS = None) -> dict:
    """Stop a service.

    Stop a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service stop <service-name>
      duploctl service stop --all
      duploctl service stop --targets service1 service2 service3
      ```

    Args:
      name (str): The name of the service to stop.
      all (bool): Boolean flag to stop all services. Defaults to False.
      targets (list[str]): List of service names to stop. Cannot be used with name or all.
      wait: Boolean flag to wait for service updates.

    Returns: 
      A summary containing services that were stopped successfully and those that encountered errors.
      
    Raises:
      DuploError: If the service could not be stopped.
    """
    return self._perform_action("Stop", name, all, targets, self.duplo.wait)

  @Command()
  def start(self,
            name: args.NAME = None,
            all: args.ALL = False,
            targets: args.TARGETS = None) -> dict:
    """Start a service.

    Start a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service start <service-name>
      duploctl service start --all
      duploctl service start --targets service1 service2 service3
      ```
    
    Args:
      name (str): The name of the service to start.
      all (bool): Boolean flag to start all services. Defaults to False.
      targets (list[str]): List of service names to start. Cannot be used with name or all.
      wait: Boolean flag to wait for service updates.

    Returns: 
      A summary containing services that were started successfully and those that encountered errors.

    Raises:
      DuploError: If the service could not be started.
    """
    return self._perform_action("start", name, all, targets, self.duplo.wait)

  @Command()
  def pods(self, 
           name: args.NAME) -> dict:
    """Get the pods for a service.
    
    Args:
      name: The name of the service to get pods for.
    Returns: 
      message: A list of pods for the service.
    Raises:
      DuploError: If the service could not be found.
    """
    def controlled_by_service(pod):
      cb = pod.get("ControlledBy", None)
      same_name = pod["Name"] == name
      if cb is None:
        return same_name
      else:
        return same_name and cb.get("QualifiedType", None) == "kubernetes:apps/v1/ReplicaSet" 
    pods = self.__pod_svc.list()
    return [ pod for pod in pods if controlled_by_service(pod) ]
  
  @Command()
  def logs(self,
           name: args.NAME) -> dict:
    """Service Logs
    
    Get the logs for a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service logs <service-name>
      ```
    Args:
      name (str): The name of the service to get logs for.
      wait (bool): Whether to wait for logs to update.
    Returns:
      message: A success message.
    """
    def show_logs():
      pods = self.pods(name)
      for pod in pods:
        self.__pod_svc.logs(pod=pod)
    if self.duplo.wait:
      try:
        while True:
          show_logs()
          time.sleep(3)
      except KeyboardInterrupt:
        pass
    else:
      show_logs()

  def current_replicaset(self, name: str):
    """Get the current replicaset for a service.
    
    Args:
      name (str): The name of the service to get replicaset for.
    Returns: 
      The current replicaset for the service.
    Raises:
      DuploError: If the service could not be found.
    """
    pods = self.pods(name)
    cb = pods[0].get("ControlledBy", None)
    if cb is None:
      return None
    return cb.get("NativeId", None)
  
  def image_from_body(self, body):
    """Get the image from a service body.
    
    Args:
      body (dict): The body of the service.
    Returns: 
      The image for the service.
    """
    tpl = body.get("Template", {})
    containers = tpl.get("Containers", [])
    for c in containers:
      if c["Name"] == body["Name"]:
        return c["Image"]
    else:
      return body.get("DockerImage", body.get("Image", None))

  def wait(self, old, updates):
    """Wait for a service to update."""
    name = old["Name"]
    cloud = old["Template"]["Cloud"]

    # track image changes
    new_img = self.image_from_body(updates)
    old_img = self.image_from_body(old)
    image_changed = (old_img != new_img) if new_img else False

    # track hpa settings and changes
    old_hpa = old.get("HPASpecs", None)
    hpa_enabled = (old_hpa is not None)
    if "HPASpecs" in updates:
      hpa_enabled = updates["HPASpecs"] is not None # it's been disabled if it's None

    # is the replica count changing and are we using the new api? 
    replica_key = "ReplicasActive" if "ReplicasActive" in old else "Replicas" # only used in wait_check
    new_replicas = updates.get("Replicas", None)
    replicas_changed = (old["Replicas"] != new_replicas) if new_replicas else False
    true_count = (replica_key == "ReplicasActive" or not hpa_enabled) # check if we can know the true count of replicas

    # Is a rollover needed? 
    new_conf = updates.get("OtherDockerConfig", None)
    conf_changed = (old["Template"].get("OtherDockerConfig", None) != new_conf) if new_conf else False
    rollover = False
    if image_changed or conf_changed:
      rollover = True
    self.duplo.logger.debug(f"""Wait for service {name}:
      The image has {'changed' if image_changed else 'not changed'}.
      The replicas have {'changed' if replicas_changed else 'not changed'}.
      The hpa is {'enabled' if hpa_enabled else 'disabled'}.
      The true count is {'known' if true_count else 'unknown'}.
      A rollover is {'required' if rollover else 'not required'}.
      """)
    def get_pod_image(pod):
      """Get the image of a pod."""
      return pod["Containers"][0]["Image"].removeprefix("docker.io/library/")
    def check_pod_faults(pod, faults):
      """Check if the pod has any faults.
      
      This is for aws and gke because the faults are at the pod level.
      """
      for f in faults:
        if f["Resource"]["Name"] == pod["InstanceId"]:
          raise DuploFailedResource(f"Pod {pod['InstanceId']} raised a fault.\n{f['Description']}")
    def check_service_faults(faults):
      """Check if the service has any faults.
      
      This is only for azure because the faults are at the service level only
      """
      for f in faults:
        if f["ResourceName"] == name:
          raise DuploFailedResource(f"Service {name} raised a fault.\n{f['Description']}")
    def running_pod(pod, faults):
      """Check if a pod is running and return 1 if it is, 0 otherwise."""
      # azure doesn't have controlled by, so we skip this check if it's not there
      cb = pod.get("ControlledBy", None) 
      # skip if the pod is not controlled by the new replicaset
      if (cloud != 2 and cb is not None) and (cb["NativeId"] == old.get("Replicaset", None) and rollover):
        return 0

      # ignore this pod if the image is the old image
      if image_changed and get_pod_image(pod) == old_img:
        return 0

      # check for aws and gke faults on pod
      if cloud != 2:
        check_pod_faults(pod, faults)

      # update total running pod count if one is running
      if ((pod["CurrentStatus"] == pod["DesiredStatus"]) and pod["DesiredStatus"] == 1):
        self.duplo.logger.info(f"Pod {pod['InstanceId']} is running")
        return 1
      return 0
    def wait_check():
      svc = self.find(name)
      replicas = svc[replica_key]

      # we are still waiting if the changed values do not appear when retrieving the service
      if (image_changed and self.image_from_body(svc) != new_img):
        raise DuploStillWaiting(f"Service {name} waiting for image update")
      if (replicas_changed and replicas != new_replicas):
        raise DuploStillWaiting(f"Service {name} waiting for replicas update")
      if (conf_changed and svc["Template"].get("OtherDockerConfig", None) != new_conf):
        raise DuploStillWaiting(f"Service {name} waiting for pod to update")
      
      # now let's start checking the pods
      pods = self.pods(name)
      faults = self.tenant_svc.faults(id=self.tenant_id)

      # check for azure faults on service
      if cloud == 2:
        check_service_faults(faults)

      # make sure all the replicas are up if we know the true replica count
      running = sum(running_pod(p, faults) for p in pods)
      if true_count and replicas != running:
        raise DuploStillWaiting(f"Service {name} waiting for pods {running}/{replicas}")
      else:
        # if this is the case, we need to hack it, this means hpa is enabled and we don't know the true replica count
        # we will check to make sure no pods in the pod list are using the old image, when none are running and at least one is up, we are done
        running_old = sum(1 for pod in pods if get_pod_image(pod) == old_img)
        min_replicas = svc.get("HPASpecs", {}).get("minReplicas", 1)
        if running_old > 0 and running < min_replicas:
          raise DuploStillWaiting(f"Service {name} waiting for pods {running}/{min_replicas} and {running_old} pods using old image to cleanup.")
    
    # send to the base class to do the waiting
    super().wait(wait_check, 3600, 11)

  def name_from_body(self, body):
    return body["Name"]

  @Command()
  def expose(self,
             name: args.NAME,
             container_port: args.CONTAINER_PORT=None,
             external_port: args.EXTERNAL_PORT=None,
             lb_type: args.LOAD_BALANCER_TYPE=None,
             protocol: args.PROTOCOL=None,
             visibility: args.LOAD_BALANCER_VISIBILITY="public",
             mode: args.LOAD_BALANCER_MODE="docker-mode",
             health_check_url: args.HEALTH_CHECK_URL=None) -> dict:
    """
    Expose a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service expose <service-name> --lb-type applicationlb --container-port 80 --external-port 80  \
                             --visibility public --mode docker-mode --health-check-url / --protocol http
      ```

    Args:
        container-port (int): The internal port of the container to expose.
        external-port (int): The external port exposed by the load balancer. This is not used for targetgrouponly or k8clusterip load balancer types.
        lb-type (str): The load balancer type. Valid options are ['applicationlb', 'k8clusterip', 'k8nodeport', 'networklb', 'targetgrouponly'].
        protocol (str): The protocol to use, based on `lb_type`
          - applicationlb: http, https
          - networklb: tcp, udp, tls
          - targetgrouponly: http, https
          - k8clusterip: tcp, udp
          - k8nodeport: tcp, udp
        visibility (str): The load balancer visibility. Valid options are 'public' or 'private'.
        mode (str): The load balancer application mode. Valid options are 'docker-mode' or 'native-app'.
        health_check_url (str): The health check URL path. This must be empty for networklb, as it does not support health check paths.
    """
    service = self.find(name)
    tenant_id = service["TenantId"]
    lb_type_map = {
      "applicationlb": 1,
      "k8clusterip": 3,
      "k8nodeport": 4,
      "networklb": 6,
      "targetgrouponly": 7
    }

    if lb_type not in lb_type_map:
      raise DuploError(f"Invalid lb_type: {lb_type}. Must be one of {list(lb_type_map.keys())}")

    payload = {
        "LbType": lb_type_map[lb_type],
        "Port": container_port,
        "ExternalPort": external_port if external_port is not None else None,
        "IsInternal": visibility == "private",
        "IsNative": mode == "native-app",
        "Protocol": protocol,
        "HealthCheckUrl": health_check_url if lb_type != "networklb" else None,
        "ReplicationControllerName": name
    }

    self.duplo.post(f"subscriptions/{tenant_id}/LBConfigurationUpdate", payload)
    return {"message": f"Successfully exposed service '{name}'"}

  @Command()
  def rollback(self,
              name: args.NAME,
              to_revision: args.TO_REVISION = None) -> dict:
    """Rollback Service

    Roll back a service to a specific revision (if provided) or the last known good state.

    Usage:
      ```sh
      duploctl service rollback <service-name>
      duploctl service rollback <service-name> --to-revision 2
      ```

    Args:
        name (str): The name of the service to roll back.
        to_revision (int, optional): The revision number to roll back to.

    Returns:
        A success message indicating the rollback status.

    Raises:
      DuploError: If the service could not be rollback.
    """
    service = self.find(name)
    tenant_id = service["TenantId"]
    api_endpoint = f"v3/subscriptions/{tenant_id}/containers/replicationController/{name}/rollback"
    if to_revision:
      api_endpoint += f"/{to_revision}"
    self.duplo.put(api_endpoint)
    return {"message": f"Successfully rolled back service '{name}'"}


import pytest
import time
from duplocloud.errors import DuploError
from .conftest import get_test_data

@pytest.fixture(scope="class")
def asg_resource(duplo):
    """Fixture to load the ASG resource and define ASG name."""
    resource = duplo.load("asg")
    resource.duplo.wait = True
    tenant = resource.tenant["AccountName"]
    asg_name = f"duploservices-{tenant}-duploctl"
    return resource, asg_name

def execute_test(func, *args, **kwargs):
    """Helper function to execute a test and handle errors."""
    try:
        return func(*args, **kwargs)
    except DuploError as e:
        pytest.fail(f"Test failed: {e}")

class TestAsg:

    @pytest.mark.integration
    @pytest.mark.dependency(name="create_asg", scope="session")
    @pytest.mark.order(1)
    def test_create_asg(self, asg_resource):
        r, asg_name = asg_resource
        body = get_test_data("asg")
        response = execute_test(r.create, body=body)
        assert response["data"] == asg_name
        time.sleep(60)

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_asg"], scope="session")
    @pytest.mark.order(2)
    def test_find_asg(self, asg_resource):
        r, asg_name = asg_resource
        asg = execute_test(r.find, asg_name)
        assert asg["FriendlyName"] == asg_name

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_asg"], scope="session")
    @pytest.mark.order(3)
    def test_update_asg(self, asg_resource):
        r, asg_name = asg_resource
        body = {"FriendlyName": asg_name, "MinSize": 2, "MaxSize": 3}
        response = execute_test(r.update, body=body)
        assert "Successfully updated asg" in response["message"]

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_asg"], scope="session")
    @pytest.mark.order(4)
    def test_list_asgs(self, asg_resource):
        r, _ = asg_resource
        asgs = execute_test(r.list)
        assert isinstance(asgs, list) and len(asgs) > 0

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_asg"], scope="session")
    @pytest.mark.order(5)
    def test_scale_asg(self, asg_resource):
        r, asg_name = asg_resource
        response = execute_test(r.scale, asg_name, min=1, max=2)
        assert "Successfully updated asg" in response["message"]

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_asg"], scope="session")
    @pytest.mark.order(6)
    def test_delete_asg(self, asg_resource):
        r, _ = asg_resource
        response = execute_test(r.delete, "duploctl")
        assert "Successfully deleted asg" in response["message"]



import pytest
import time
from duplocloud.errors import DuploError, DuploFailedResource, DuploStillWaiting
from .conftest import get_test_data

@pytest.fixture(scope="class")
def job_resource(duplo):
    """Fixture to load the Job resource and define Job name."""
    resource = duplo.load("job")
    resource.duplo.wait = True
    job_name = f"duploctl"
    return resource, job_name

def execute_test(func, *args, **kwargs):
    """Helper function to execute a test and handle errors."""
    try:
        return func(*args, **kwargs)
    except (DuploError, DuploFailedResource, DuploStillWaiting) as e:
        pytest.fail(f"Test failed: {e}")

class TestJob:

    @pytest.mark.integration
    @pytest.mark.dependency(name="create_job", scope="session")
    @pytest.mark.order(1)
    def test_create_job(self, job_resource):
        """Test creating a new job."""
        r, job_name = job_resource
        body = get_test_data("job")
        response = execute_test(r.create, body=body)
        assert "ran successfully" in response["message"]
        time.sleep(30)  # Allow time for pods to be created

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_job"], scope="session")
    @pytest.mark.order(2)
    def test_find_job(self, job_resource):
        """Test finding a specific job."""
        r, job_name = job_resource
        job = execute_test(r.find, job_name)
        assert job["metadata"]["name"] == job_name

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_job"], scope="session")
    @pytest.mark.order(3)
    def test_list_jobs(self, job_resource):
        """Test listing all jobs."""
        r, _ = job_resource
        jobs = execute_test(r.list)
        assert isinstance(jobs, list) and len(jobs) > 0

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_job"], scope="session")
    @pytest.mark.order(4)
    def test_get_pods(self, job_resource):
        """Test getting pods for a job."""
        r, job_name = job_resource
        pods = execute_test(r.pods, job_name)
        assert isinstance(pods, list)
        if len(pods) > 0:
            assert pods[0]["ControlledBy"]["QualifiedType"] == "kubernetes:batch/v1/Job"
            assert pods[0]["Name"] == job_name

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_job"], scope="session")
    @pytest.mark.order(5)
    def test_delete_job(self, job_resource):
        """Test deleting a job."""
        r, job_name = job_resource
        response = execute_test(r.delete, job_name)
        assert "deleted" in response["message"].lower()



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.errors import DuploError
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("batch_queue")
class DuploBatchQueue(DuploTenantResourceV3):
  """Manage AWS Batch Job Resources

  Run batch jobs as a managed service on AWS infrastructure. 

  Read more docs here: 
  https://docs.duplocloud.com/docs/overview/aws-services/batch
  """

  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, 
                     slug="aws/batchJobQueue",
                     prefixed=True)

  def name_from_body(self, body):
    return body["JobQueueName"]
  
  @Command()
  def create(self, body: args.BODY) -> dict:
    """Create a Batch Job Queue.

    Creates a new Batch Job Queue with the specified configuration.

    Usage: Basic CLI Use
      ```sh
      duploctl batch_queue create -f batch_queue.yaml
      ```

    Args:
      body: The configuration for the Batch Job Queue.

    Returns:
      dict: The created Batch Job Queue object.
    """
    arn = super().create(body)
    return {
      "Message": "Batch Job Queue created successfully.",
      "JobQueueArn": arn
    }

  @Command()
  def find(self, 
           name: args.NAME) -> dict:
    """Find a Single Batch Job Queue by name.

    Usage: cli usage
      ```sh
      duploctl batch_queue find <name>
      ```

    Args:
      name: The name of the Batch Job Queue to find.

    Returns: 
      resource: The Batch Job Queue object.
    """
    n = self.prefixed_name(name)
    queues = self.list()
    for q in queues:
      if self.name_from_body(q) == n:
        return q
    raise DuploError(f"Batch Job Queue '{name}' not found", 404)

  @Command()
  def disable(self, 
             name: args.NAME) -> dict:
    """Disable a Batch Compute Environment.

    Usage: cli usage
      ```sh
      duploctl batch_compute delete <name>
      ```
    
    Args:
      name: The name of the Batch Compute Environment to delete.

    Returns: 
      message: A success message.

    Raises:
      DuploError: If the Batch Compute Environment could not be found or deleted. 
    """
    n = self.prefixed_name(name)
    endpoint = f"{self.endpoint()}Disable/{n}"
    self.duplo.delete(endpoint)
    return {
      "message": f"{self.slug}/{name} disabled"
    }


from http.server import HTTPServer, SimpleHTTPRequestHandler
from .errors import DuploError
import threading
import time
import webbrowser

class TokenCallbackHandler(SimpleHTTPRequestHandler):

  def do_POST(self):
    """Do Post
    
    The post request to receive the token. The token is read from the request body and stored in the server instance.
    """
    content_length = int(self.headers['Content-Length'])
    post_data = self.rfile.read(content_length)
    
    # Send response back to client
    self.send_response(200)
    self.end_headers()
    self.wfile.write(b'"done"')
    token = post_data.decode('utf-8')
    self.server.token = token

  def do_OPTIONS(self):
    """Do Options
    
    The preflight request for CORS.
    """
    self.send_response(200, "ok")
    self.end_headers()

  def end_headers(self):
    """End Headers
    
    Override the end headers to add the cors headers and prevent caching.
    """
    self.send_header('Access-Control-Allow-Origin', self.server.host)
    self.send_header('Access-Control-Allow-Methods', 'POST, OPTIONS')
    self.send_header('Access-Control-Allow-Headers', '*')
    self.send_header('Cache-Control', 'no-store, no-cache, must-revalidate')
    return super(TokenCallbackHandler, self).end_headers()

  def log_message(self, format, *args):
    # Override to prevent printing any log messages
    pass

class TokenServer(HTTPServer):
  def __init__(self, host: str, timeout=60):
    """TokenServer
    
    A simple HTTP server to receive a token from a callback. The bind host is always empty for localhost and the port is always 0 to get a random port. The server is started in a separate thread and the token is received in the main thread. The given host is the only host that is allowed to send a token and this is enforced in the allow origin cors header.

    Args:
      host: The host to receive the callbcack from.
      timeout: The timeout to wait for a token.

    """
    self.token = None
    self.host = host
    self.timeout = timeout
    super().__init__(('', 0), TokenCallbackHandler, True)

  def serve_token(self):
    """Serve Token
    
    Start the server and wait for a token. This is a blocking call and will wait for the token to be received from the callback or the timeout to expire. If the timeout expires, a 403 error is raised.
    """
    st = threading.Thread(target=self.serve_forever)
    wt = threading.Thread(target=self.wait_for_token)
    st.start()
    wt.start()
    wt.join(timeout=self.timeout)
    st.join()
    if not self.token:
      raise DuploError("Failed to get token", 403)
    return self.token

  def wait_for_token(self):
    """Wait for Token
    
    Simply waits for the token to be set by the handler or the timeout to expire. 
    Ultimately the server is shutdown so no more threads are used. 
    """
    i = 0
    while not self.token and i < self.timeout:
      time.sleep(1)
      i += 1
    self.shutdown()

  def open_callback(self, page: str, browser=None):
    """Open Callback

    Opens the configured hosts callback page in the browser. 

    Args:
      page: The page to open in the browser.
      browser: The browser to use. If not specified, the default browser is used.
    """
    url = f"{self.host}/{page}"
    wb = webbrowser if not browser else webbrowser.get(browser)
    wb.open(url, new=0, autoraise=True)



import pytest
from duplocloud.errors import DuploError
from .conftest import get_test_data

@pytest.fixture(scope="class")
def secret_resource(duplo, request):
    """Fixture to load the Kubernetes Secret resource and define the secret name."""
    resource = duplo.load("secret")
    request.cls.secret_name = f"duploctl"
    return resource

def execute_test(func, *args, **kwargs):
    """Helper function to execute a test and handle errors."""
    try:
        return func(*args, **kwargs)
    except DuploError as e:
        pytest.fail(f"Test failed: {e}")

class TestSecret:
    """Integration tests for Kubernetes Secrets."""

    @pytest.mark.integration
    @pytest.mark.dependency(name="create_secret", scope="session")
    @pytest.mark.order(1)
    def test_create_secret(self, secret_resource):
        """Test creating a Kubernetes secret."""
        body = get_test_data("secret")
        execute_test(secret_resource.create, name=self.secret_name, body=body)

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_secret"], scope="session")
    @pytest.mark.order(2)
    def test_find_secret(self, secret_resource):
        """Test finding the created Kubernetes secret."""
        secret = execute_test(secret_resource.find, self.secret_name)
        assert secret["SecretName"] == self.secret_name
        assert "SecretData" in secret
        assert "username" in secret["SecretData"]
        assert "password" in secret["SecretData"]
        assert secret["SecretData"]["username"] == "admin"
        assert secret["SecretData"]["password"] == "secret123"

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_secret"], scope="session")
    @pytest.mark.order(3)
    def test_update_secret(self, secret_resource):
        """Test updating a Kubernetes secret."""
        new_data = {"api_key": "xyz789","username": "superadmin"}
        execute_test(secret_resource.update,name=self.secret_name,data=new_data)
        # Verify the update
        updated_secret = execute_test(secret_resource.find, self.secret_name)
        assert updated_secret["SecretData"]["api_key"] == "xyz789"
        assert updated_secret["SecretData"]["username"] == "superadmin"
        assert updated_secret["SecretData"]["password"] == "secret123"
        # Test update using patches
        patches = [
            {'op': 'add', 'path': '/SecretData/new_key', 'value': 'new_value'},
            {'op': 'remove', 'path': '/SecretData/password', 'value': None}
        ]
        execute_test(secret_resource.update, name=self.secret_name, patches=patches)
        # Verify patches
        patched_secret = execute_test(secret_resource.find, self.secret_name)
        assert patched_secret["SecretData"]["new_key"] == "new_value"
        assert "password" not in patched_secret["SecretData"]

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_secret"], scope="session")
    @pytest.mark.order(4)
    def test_delete_secret(self, secret_resource):
        """Test deleting a Kubernetes secret."""
        response = execute_test(secret_resource.delete, self.secret_name)
        assert response["message"] == f"Successfully deleted secret 'duploctl'"


from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.errors import DuploError, DuploStillWaiting
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("ecs")
class DuploEcsService(DuploTenantResourceV2):

  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)

  @Command()
  def list_services(self):
    """Retrieve a list of all ECS services in a tenant."""
    tenant_id = self.tenant["TenantId"]
    url = f"subscriptions/{tenant_id}/GetEcsServices"
    response = self.duplo.get(url)
    return response.json()

  @Command()
  def find_service_family(self,
           name: args.NAME):
    """Find an ECS Services task definition family by name.

    Args:
      name (str): The name of the ECS task definition to find.
    Returns:
      The ECS task definition object.
    Raises:
      DuploError: If the ECS task definition could not be found.
    """
    tenant_id = self.tenant["TenantId"]
    path = f"v3/subscriptions/{tenant_id}/aws/ecs/service/taskDefFamily/{name}"
    response = self.duplo.get(path)
    return response.json()

  @Command()
  def delete_service(self,
                     name: args.NAME) -> dict:
    """Delete an ECS service.

    Args:
      name: The name of the ECS service to delete.
    Returns:
      message: A message indicating the service has been deleted.
    """
    tenant_id = self.tenant["TenantId"]
    path = f"subscriptions/{tenant_id}/DeleteEcsService/{name}"
    response = self.duplo.post(path, {})
    return response.json()

  @Command()
  def list_task_def_family(self) -> dict:
    """List ECS Task Definitions

    Retrieve a list of all ECS task definitions in a tenant.

    Example:
      CLI usage
      ```sh
      duploctl ecs list_definitions
      ```

    Returns:
      task_def_family: A list of ECS task definitions.
    """
    tenant_id = self.tenant["TenantId"]
    path = f"v3/subscriptions/{tenant_id}/aws/ecs/taskDefFamily"
    response = self.duplo.get(path)
    return response.json()

  @Command()
  def find_def(self,
               name: args.NAME):
    """Find a ECS task definition by name.

    Args:
      name (str): The name of the ECS task definition to find.
    Returns:
      The ECS task definition object.
    Raises:
      DuploError: If the ECS task definition could not be found.
    """
    name = self.prefixed_name(name)
    tdf = self.find_task_def_family(name)
    arn = tdf["VersionArns"][-1]
    return self.find_def_by_arn(arn)

  @Command()
  def find_def_by_arn(self,
                      arn: args.ARN):
    """Find a ECS task definition by ARN.

    Args:
      arn (str): The ARN of the ECS task definition to find.
    Returns:
      The ECS task definition object.
    Raises:
      DuploError: If the ECS task definition could not be found.
    """
    path = self.endpoint("FindEcsTaskDefinition")
    response = self.duplo.post(path, {"Arn": arn})
    return response.json()

  @Command()
  def find_task_def_family(self,
                           name: args.NAME):
    """Find a ECS task definition by name.

    Args:
      name (str): The name of the ECS task definition to find.
    Returns:
      The ECS task definition object.
    Raises:
      DuploError: If the ECS task definition could not be found.
    """
    name = self.prefixed_name(name)
    tenant_id = self.tenant["TenantId"]
    path = f"v3/subscriptions/{tenant_id}/aws/ecs/taskDefFamily/{name}"
    response = self.duplo.get(path)
    return response.json()

  @Command()
  def update_service(self,
             body: args.BODY):
    """Update an ECS service.

    Args:
      body (dict): The updated ECS service object.
    Returns:
      The updated ECS object.
    Raises:
      DuploError: If the ECS service could not be updated.
    """
    path = self.endpoint("UpdateEcsService")
    self.duplo.post(path, body)
    return {"message": "ECS Service updated"}

  @Command()
  def update_taskdef(self,
                     body: args.BODY):
    """Update an ECS task definition.

    Args:
      body (dict): The updated ECS task definition object.
    Returns:
      The updated ECS object.
    Raises:
      DuploError: If the ECS task definition could not be updated.
    """
    path = self.endpoint("UpdateEcsTaskDefinition")
    b = self.__ecs_task_def_body(body)
    response = self.duplo.post(path, b)
    return {"arn": response.json()}

  @Command()
  def update_image(self,
                   name: args.NAME,
                   image: args.IMAGE = None,
                   container_image: args.CONTAINER_IMAGE = None) -> dict:
    """Update the image for an ECS service container.

    Usage: Basic CLI Use
      ```sh
        duploctl ecs update_image <service-name> <service-image>
      ```
      ```sh
        duploctl ecs update_image <service-name> --container-image <container-name> <container-image>
      ```

    Args:
      name: The name of the ECS service to update.
      image: The new image to use for the container.
      container-image: A list of key-value pairs to set as container image.

    Returns:
      dict: A dictionary containing a message about the update status.

    Raises:
      DuploError: If the ECS service could not be updated.
    """
    name = self.prefixed_name(name)
    tdf = self.find_def(name)
    if container_image:
      container_updates = dict(container_image)
      for container_def in tdf["ContainerDefinitions"]:
        if container_def["Name"] in container_updates:
          container_def["Image"] = container_updates[container_def["Name"]]
    if image:
      tdf["ContainerDefinitions"][0]["Image"] = image
    arn = self.update_taskdef(tdf)["arn"]
    msg = "Updating a task definition and its corresponding service."
    svc = None
    try:
      svcFam = self.find_service_family(name)
      svc = svcFam["DuploEcsService"]
      svc["TaskDefinition"] = arn
    except DuploError:
      msg = "No Service Configured, only the definition is updated."
    # run update here so the errors bubble up correctly
    if svc:
      self.update_service(svc)
      if self.duplo.wait:
        self.wait(lambda: self.wait_on_task(name))
    return {
      "message": msg
    }

  def __ecs_task_def_body(self, task_def):
    containers = [
      self.__ecs_container_update_body(c)
      for c in task_def.get("ContainerDefinitions", [])
    ]
    result = {
      "Family": task_def["Family"],
      "InferenceAccelerators": task_def.get("InferenceAccelerators", []),
      "NetworkMode": task_def.get("NetworkMode", {}),
      "ContainerDefinitions": containers,
      "RuntimePlatform": task_def.get("RuntimePlatform", {})
    }
    if task_def.get("Cpu") not in (None, 0):
      result["Cpu"] = task_def["Cpu"]
    if task_def.get("Memory") not in (None, 0):
      result["Memory"] = task_def["Memory"]
    if task_def.get("MemoryReservation") not in (None, 0):
      result["MemoryReservation"] = task_def["MemoryReservation"]
    return result


  def __ecs_container_update_body(self, container_def):
    update_body = {
        "Essential": container_def.get("Essential"),
        "Image": container_def.get("Image") ,
        "Name": container_def.get("Name") ,
        "PortMappings": container_def.get("PortMappings", []) ,
        "Environment": container_def.get("Environment", {}) ,
        "Command": container_def.get("Command", {}) ,
        "Secrets": container_def.get("Secrets", {}) ,
        "EnvironmentFiles": container_def.get("EnvironmentFiles", {})
    }

    # Add LogConfiguration only if it exists in container_def
    if "LogConfiguration" in container_def:
        update_body["LogConfiguration"] = container_def["LogConfiguration"]
    # Add FirelensConfiguration only if it exists in container_def
    if "FirelensConfiguration" in container_def:
        update_body["FirelensConfiguration"] = container_def["FirelensConfiguration"]
    if container_def.get("Cpu") not in (None, 0):
        update_body["Cpu"] = container_def["Cpu"]
    if container_def.get("Memory") not in (None, 0):
        update_body["Memory"] = container_def["Memory"]
    if container_def.get("MemoryReservation") not in (None, 0):
        update_body["MemoryReservation"] = container_def["MemoryReservation"]
    return update_body

  @Command()
  def list_tasks(self,
                 name: args.NAME) -> list:
    """List tasks for an ECS service.
    """
    tenant_id = self.tenant["TenantId"]
    path = f"v3/subscriptions/{tenant_id}/aws/ecsTasks/{name}"
    res = self.duplo.get(path)
    return res.json()

  @Command()
  def run_task(self,
               name: args.NAME,
               replicas: args.REPLICAS) -> dict:
    """Run a task for an ECS service."

    Execute a task based on some definition.

    Args:
      name: The name of the ECS service to run the task for.
      replicas: The number of replicas to run.
      wait: Whether to wait for the task to complete.
    Returns:
      message: A message indicating the task has been run.
    """
    td = self.find_def(name)
    tenant_id = self.tenant["TenantId"]
    path = f"v3/subscriptions/{tenant_id}/aws/runEcsTask"
    body = {
      "TaskDefinition": td["TaskDefinitionArn"],
      "Count": replicas if replicas else 1
    }
    res = self.duplo.post(path, body)
    if self.duplo.wait:
      self.wait(lambda: self.wait_on_task(name))
    return res.json()

  def wait_on_task(self,
                   name: str) -> None:
    """Wait for an ECS task to complete."""
    tasks = self.list_tasks(name)
    # filter the tasks down to any where the DesiredStatus and LastStatus are different
    running_tasks = [t for t in tasks if t["DesiredStatus"] != t["LastStatus"]]
    if len(running_tasks) > 0:
      raise DuploStillWaiting(f"Service {name} waiting for replicas update")



from duplocloud.client import DuploClient
from duplocloud.errors import DuploError
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("configmap")
class DuploConfigMap(DuploTenantResourceV3):
  """Kubernetes ConfigMaps
  
  This class offers methods to manage Kubernetes ConfigMaps within DuploCloud.

  See more details at:
  https://docs.duplocloud.com/docs/kubernetes-overview/configs-and-secrets
  """

  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "k8s/configmap")

  @Command()
  def create(self, 
             name: args.NAME=None,
             body: args.BODY=None,
             data: args.DATAMAP=None,
             dryrun: args.DRYRUN=False,) -> dict:
    """Create a Configmap resource.

    Creates a new kubernetes ConfigMap resource with the specified metadata and data entries.

    Usage: CLI Usage
      ```sh
      duploctl configmap create -f configmap.yaml
      ```
      Contents of the `configmap.yaml` file
      ```yaml
      --8<-- "src/tests/data/configmap.yaml"
      ```

    Example: Create a ConfigMap using a one-liner.
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/configmap.yaml"
      \"\"\" | duploctl configmap create -f -
      ```

    Example: Create a ConfigMap using a file.
      ```sh
      duploctl configmap create -f configmap.yaml
      ```

    Example: Create a ConfigMap by specifying key-value pairs as literals.
      ```sh
      duploctl configmap create <configmap-name> --from-literal Key1="Val1" --from-literal Key2="Val2"
      ```

    Example: Create a ConfigMap from a file.
      ```sh
      duploctl configmap create <configmap-name> --from-file config-map.txt
      ```

    Args:
      name: Name of the ConfigMap. Required if `body` is not provided.
      body: The complete ConfigMap resource definition.
      data: Data to merge into the ConfigMap.
      dryrun: Do not submit any changes to the server.

    Returns:
      message: The created resource or success message.

    Raises:
      DuploError: If the Configmap could not be created.
    """
    if not name and not body:
      raise DuploError("Name is required when body is not provided")
    if not body:
      body = {}
    # make sure the body has a metadata key
    if 'metadata' not in body:
      body['metadata'] = {}
    # also make sure the data key is present
    if 'data' not in body:
      body['data'] = {}
    if name:
      body['metadata']['name'] = name
    if data:
      body['data'].update(data)
    if dryrun:
      return body
    else:
      return super().create(body)

  @Command()
  def update(self,
             name: args.NAME,
             body: args.BODY=None,
             data: args.DATAMAP=None,
             patches: args.PATCHES = None,
             dryrun: args.DRYRUN=False) -> dict:
    """Updates a ConfigMap resource.

    This function allows you to modify the contents of a ConfigMap without deleting
    or recreating it.

    Usage: CLI Usage
      ```sh
      duploctl configmap update -f configmap.yaml
      ```
      Contents of the `configmap.yaml` file
      ```yaml
      --8<-- "src/tests/data/configmap.yaml"
      ```

    Example: Update configmap using a one-liner.
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/configmap.yaml"
      \"\"\" | duploctl configmap update -f -
      ```

    Example: Add new key in the configmap.
      ```sh
      duploctl configmap update <configmap-name> --add /data/NewKey NewValue
      ```

    Example: Update existing key from the configmap.
      ```sh
      duploctl configmap update <configmap-name> --replace /data/ExistingKey NewValue
      ```

    Example: Delete existing key from the configmap.
      ```sh
      duploctl configmap update <configmap-name> --remove /data/ExistingKey
      ```

    Example: Update a ConfigMap by specifying key-value pairs as literals.
      ```sh
      duploctl configmap update <configmap-name> --from-literal Key1="Val1" --from-literal Key2="Val2"
      ```

    Example: Update a ConfigMap from a file.
      ```sh
      duploctl configmap update <configmap-name> --from-file config-map.txt
      ```

    Args:
      name: Name of the ConfigMap. Required if `body` is not provided.
      body: The complete ConfigMap resource definition.
      data: Data to merge into the ConfigMap.
      patches: A list of JSON patches as args to apply to the service.
        The options are `--add`, `--remove`, `--replace`, `--move`, and `--copy`.
        Then followed by `<path>` and `<value>` for `--add`, `--replace`, and `--test`.
      dryrun (bool, optional): If True, return the modified ConfigMap without applying changes.

    Returns:
      message: The updated ConfigMap or a success message.

    Raises:
      DuploError: If the ConfigMap update fails.
    """
    if data:
      if not name:
        raise DuploError("Name is required when body is not provided")
      body = self.find(name)
      body.setdefault('data', {}).update(data or {})
    return body if dryrun else super().update(name=name, body=body, patches=patches)

  @Command()
  def find(self,
           name: args.NAME) -> dict:
    """Find a ConfigMap.

    Retrieve details of a specific ConfigMap by name

    Usage: cli usage
      ```sh
      duploctl configmap find <name>
      ```

    Args:
      name: The name of the ConfigMap to find.

    Returns:
      message: The resource content or success message.

    Raises:
      DuploError: ConfigMap not found.
    """
    try:
      response = self.duplo.get(self.endpoint(name))
    except DuploError as e:
      raise DuploError(f"Failed to find ConfigMap '{name}': {str(e)}")
    return response.json()

  @Command()
  def delete(self,
             name: args.NAME) -> dict:
    """Delete ConfigMap

    Deletes the specified ConfigMap by name.

    Usage: cli
      ```sh
      duploctl configmap delete <name>
      ```

    Args:
      name: The name of a ConfigMap to delete.

    Returns:
      message: Returns a success message if deleted successfully; otherwise, an error.
    """
    super().delete(name)
    return {
      "message": f"Successfully deleted configmap '{name}'"
    }



#!/usr/bin/env python3
import argparse
import sys
import os
import logging
import importlib.metadata as meta
HERE = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(HERE))
from gha import GithubRepo
from project import Project, CHANGELOG

logging.basicConfig(level=logging.INFO)

class Versionizer:
  def __init__(self, action, token=None):
    self.logger = logging.getLogger('versionizer')
    self.gha = GithubRepo(token)
    self.project = Project()
    self.last_version = str(self.project.latest_tag)
    self.ref = str(self.project.ref)
    self.action = action
    self.pushed = False
    self.changelog = None
    self.notes = None
    self.__next_version = None

  @property 
  def next_version(self):
    if not self.__next_version:
      self.__next_version = str(self.project.bump_version(self.action))
    return self.__next_version

  def build_release_notes(self, save=True):
    self.logger.info("Building release notes")
    v = self.next_version
    lv = self.last_version
    cl_notes = self.project.release_notes()
    pr_notes = self.gha.generate_release_notes(f"v{v}", f"v{lv}", self.ref)
    inst_notes = self.project.install_notes(v)
    notes = "\n".join([
      cl_notes, 
      pr_notes["body"], 
      "\n## Installation",
      inst_notes])
    if save:
      self.project.dist_file('notes.md', notes)
    self.notes = notes
  
  def reset_changelog(self, save=True):
    self.logger.info(f"Resetting changelog")
    c = self.project.reset_changelog(self.next_version)
    if save:
      self.project.dist_file(CHANGELOG, c)
    self.changelog = c

  def publish(self):
    if not self.pushed:
      self.logger.info(f"Publishing new version {self.next_version}")
      self.gha.publish(f"v{self.next_version}", CHANGELOG, self.changelog)
      self.pushed = True

  def save_github_output(self):
    """Save the tag and version in github output file"""
    tag = f"v{self.next_version}" if self.pushed else self.ref
    version = self.next_version if self.pushed else meta.version('duplocloud-client')
    outputs = os.environ.get('GITHUB_OUTPUT', './.github/output')
    self.logger.info(f"Saving Outputs: tag={tag} version={version}")
    with open(outputs, 'a') as f:
      f.write(f"tag={tag}\n")
      f.write(f"version={version}\n")

if __name__ == '__main__':
  parser = argparse.ArgumentParser(
    prog='version-bumper',
    description='A duploctl version bumper.',
  )
  parser.add_argument('--action', 
                      type=str, 
                      help='The type of version bump to perform.', 
                      choices=['major', 'minor', 'patch'], 
                      default="patch")
  parser.add_argument('--push', 
                      type=str, 
                      help='Push to remote?', 
                      default="false")
  parser.add_argument('--token', 
                      type=str, 
                      help='Github token', 
                      default=os.getenv('GITHUB_TOKEN', None))
  args = parser.parse_args()
  v = Versionizer(args.action, args.token)
  v.build_release_notes()
  v.reset_changelog()
  if args.push == "true":
    v.publish()
  v.save_github_output()



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.errors import DuploError
from duplocloud.commander import Resource, Command
import duplocloud.args as args

@Resource("ingress")
class DuploIngress(DuploTenantResourceV3):
  """Kubernetes Ingress

  This class offers methods to manage Kubernetes Ingress within DuploCloud.
  
  See more details at: https://docs.duplocloud.com/docs/kubernetes-overview/ingress-loadbalancer
  """
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "k8s/ingress")

  def name_from_body(self, body):
    return body["name"]

  @Command()
  def create(self,
             body: args.BODY) -> dict:
    """Create an Ingress resource.

    Creates a new Ingress resource with the specified metadata and data entries.

    Usage: CLI Usage
      ```sh
      duploctl ingress create -f ingress.yaml
      ```
      Contents of the `ingress.yaml` file
      ```yaml
      --8<-- "src/tests/data/ingress.yaml"
      ```

    Example: Create an Ingress using a one-liner.
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/ingress.yaml"
      \"\"\" | duploctl ingress create -f -
      ```

    Example: Create an Ingress using a file.
      ```sh
      duploctl ingress create -f ingress.yaml
      ```

    Args:
      body: The complete Ingress resource definition including name, rules, and other configuration.

    Returns:
      message: The created resource or success message.

    Raises:
      DuploError: If the Ingress could not be created due to invalid configuration or API errors.
    """
    name = self.name_from_body(body)
    response = self.duplo.post(self.endpoint(), body)
    return {
      "message": f"Successfully Created an Ingress '{name}'",
      "data": response.json()
    }

  @Command()
  def update(self,
             name: args.NAME,
             body: args.BODY = None,
             patches: args.PATCHES = None) -> dict:
    """Update an Ingress resource.

    Update an existing Ingress resource with the specified metadata and data entries.
    The update can be performed either by providing a complete resource definition or
    by applying JSON patches to modify specific fields.

    Usage: CLI Usage
      ```sh
      duploctl ingress update -f ingress.yaml
      ```
      Contents of the `ingress.yaml` file
      ```yaml
      --8<-- "src/tests/data/ingress.yaml"
      ```

    Example: Update ingress using a one-liner.
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/ingress.yaml"
      \"\"\" | duploctl ingress update -f -
      ```

    Example: Update dnsPrefix for an ingress.
      ```sh
      duploctl ingress update <ingress-name> --replace /lbConfig/dnsPrefix <value>
      ```

    Example: Update port of a rule for an ingress.
      ```sh
      duploctl ingress update <ingress-name> --replace /rules/0/port <port>
      ```

    Example: Update ingress by adding an additional rule.
      ```sh
      duploctl ingress update <ingress-name> --add /rules/- '{"path":"/","pathType":"Prefix","serviceName":"<service-name>","port":80,"host":"<host>","portName":null}'
      ```

    Example: Update ingress by removing a rule.
      ```sh
      duploctl ingress update <ingress-name> --remove /rules/0
      ```

    Args:
      name: The name of the Ingress resource to update. Required if `body` is not provided.
      body: The complete Ingress resource definition with updated configuration.
      patches: A list of JSON patches to apply to the Ingress resource.
        The options are `--add`, `--remove`, `--replace`, `--move`, and `--copy`.
        Then followed by `<path>` and `<value>` for `--add`, `--replace`, and `--test`.

    Returns:
      message: The created resource or success message.

    Raises:
      DuploError: If the Ingress could not be updated.
    """
    if name and body:
      if 'name' not in body or body['name'] != name:
        raise DuploError("Provided 'name' must match 'name' in the body.")
    if body is None:
      body = self.find(name)
    response = super().update(name=name, body=body, patches=patches)
    return {
      "message": f"Successfully Updated an Ingress '{name}'",
      "data": response
    }



from duplocloud.client import DuploClient
from duplocloud.errors import DuploFailedResource, DuploStillWaiting
from duplocloud.resource import DuploResource
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("infrastructure")
class DuploInfrastructure(DuploResource):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)

  @Command()
  def eks_config(self,
                 planId: args.PLAN = None):
    """Retrieve eks session credentials for current user."""
    res = self.duplo.get(f"v3/admin/plans/{planId}/k8sClusterConfig")
    return res.json()
  
  @Command()
  def list(self):
    """Retrieve a list of all infrastructures in the Duplo system."""
    response = self.duplo.get("adminproxy/GetInfrastructureConfigs/true")
    return response.json()
  
  @Command()
  def find(self, 
           name: args.NAME):
    """Find an infrastructure by name."""
    response = self.duplo.get(f"adminproxy/GetInfrastructureConfig/{name}")
    return response.json()
  
  @Command()
  def create(self, 
             body: args.BODY):
    """Create a new infrastructure."""
    status = None
    name = body["Name"]
    def wait_check():
      nonlocal status
      i = self.find(name)
      s = i.get("ProvisioningStatus", "submitted")
      if status != s:
        self.duplo.logger.warn(f"Infrastructure '{name}' - {s}")
        status = s
      if s != "Complete":
        # stop waiting if the status contains failed
        if "Failed" in s:
          raise DuploFailedResource(f"Infrastructure '{name} - {s}'")
        raise DuploStillWaiting(f"Infrastructure '{name}' is waiting for status Complete")
    self.duplo.post("adminproxy/CreateInfrastructureConfig", body)
    if self.duplo.wait:
      self.wait(wait_check, 1800, 20)
    return {
      "message": f"Infrastructure '{body['Name']}' created"
    }
  
  @Command()
  def delete(self,
             name: args.NAME):
    """Delete an infrastructure."""
    self.duplo.post(f"adminproxy/DeleteInfrastructureConfig/{name}", None)
    return {
      "message": f"Infrastructure '{name}' deleted"
    }

  @Command()
  def faults(self, 
             name: args.NAME):
    """Retrieve a list of all infrastructure faults in the Duplo system."""
    response = self.duplo.get("adminproxy/GetAllFaults")
    faults = response.json()
    response = self.duplo.get("admin/GetAllFaults")
    faults += response.json()
    return faults



from typing import Any
import requests

class GithubRepo:
  def __init__(self, token=None, repo_name="duploctl"):
    self.url = f"https://api.github.com/repos/duplocloud/{repo_name}"
    self.headers = {
      "Accept": "application/vnd.github.v3+json",
      "Authorization": f"token {token}",
    }

  def publish(self, tag, file, content) -> Any:
    parent = self.get_base_commit()
    base_tree = parent["object"]["sha"]
    tree = self.create_tree(base_tree, file, content)
    commit = self.create_commit(base_tree, tree, f"Bump version to {tag}")
    self.update_main(commit)
    self.create_tag(tag, commit)

  def get_base_commit(self):
    r = requests.get(f"{self.url}/git/refs/heads/main", headers=self.headers)
    return r.json()

  def create_tree(self, base_tree, file, content):
    r = requests.post(f"{self.url}/git/trees", headers=self.headers, json={
      "base_tree": str(base_tree),
      "tree": [{
        "path": file,
        "mode": "100644",
        "type": "blob",
        "content": content
      }]
    })
    return r.json()

  def create_commit(self, base_tree, tree, message):
    r = requests.post(f"{self.url}/git/commits", headers=self.headers, json={
      "message": message,
      "tree": tree["sha"],
      "parents": [base_tree],
    })
    return r.json()
  
  def update_main(self, commit):
    r = requests.patch(f"{self.url}/git/refs/heads/main", headers=self.headers, json={
      "sha": commit["sha"]
    })
    return r.json()
  
  def create_tag(self, tag, commit):
    r = requests.post(f"{self.url}/git/refs", headers=self.headers, json={
      "ref": f"refs/tags/{tag}",
      "sha": commit["sha"]
    })
    return r.json()
  
  def generate_release_notes(self, tag_name, previous_tag_name=None, target_commitish=None):
    body = {
      "tag_name": tag_name
    }
    if previous_tag_name:
      body["previous_tag_name"] = previous_tag_name
    if target_commitish:
      body["target_commitish"] = target_commitish
    r = requests.post(f"{self.url}/releases/generate-notes", headers=self.headers, json=body)
    return r.json()



"""
This module contains the customizations to the Argparse library. 
"""
from typing import NewType, Any, List
import argparse
import yaml
import json 
import os
from .errors import DuploError

class Arg(NewType):
  """Duplo ArgType
  
  A custom type for defining arguments in Duplo commands as annotations. 
  This extends the NewType class so type hinting will work as expected and we get truly new types. 
  Each instance of this class will be used to define a command line argument for a function.
  The values contained are the values needed for the argparse.ArgumentParser.add_argument method.

  Example:
    Make a reusable arg type called `foo`
    ```python
    FOO = Arg("foo", "-f", help="A foo arg")
    ```
  """
  def __init__(self, 
              name: str, 
              *flags: List[str], 
              type: Any=str, 
              action: str=None, 
              nargs: str=None, 
              const: str=None, 
              default=None, 
              choices=None, 
              required: bool=None, 
              help: str=None, 
              metavar: tuple=None, 
              dest: str=None,
              version: bool=None,
              env: str=None):
    """Initialize ArgType

    Args:
      name: The name of the argument.
      type: The type of the argument.
      flags: The flag to use for the argument.
      action: The action to use for the argument.
      nargs: The number of arguments to use for the argument.
      const: The constant to use for the argument.
      default (any): The default value to use for the argument. This can be overriden by the default value of the arg in the function definition.
      choices (List[any]): The choices to use for the argument.
      required: Whether the argument is required.
      help: The help text to use for the argument.
      metavar: The metavar to use for the argument.
      dest: The destination to use for the argument.
      version: The version to use for the special version arg.
      env: The environment variable to use for the argument.
    """
    super().__init__(name, type)
    self.attributes = {}
    self.env = env
    self.__flags = flags
    self.__default = default
    # if action is not a string, then type is allowed
    if not isinstance(action, str):
      self.set_attribute("type", type)
    self.set_attribute("action", action)
    self.set_attribute("nargs", nargs)
    self.set_attribute("const", const)
    self.set_attribute("choices", choices)
    self.set_attribute("required", required)
    self.set_attribute("help", help)
    self.set_attribute("metavar", metavar)
    self.set_attribute("dest", dest)
    self.set_attribute("version", version)
  def set_attribute(self, key, value):
    if value is not None:
      self.attributes[key] = value
  @property
  def flags(self):
    if self.positional:
      # return the name
      return [self.__name__]
    return [f"--{self.__name__}", *self.__flags]
  @property
  def positional(self):
    return len(self.__flags) == 0
  
  @property
  def default(self):
    if self.env:
      return os.getenv(self.env, self.__default)
    else:
      return self.__default
  
  @property 
  def type_name(self):
    # t = self.attributes.get("type", self.__supertype__)
    t = self.__supertype__
    return getattr(t, "__name__", str(t))
  
  def __str__(self):
    return self.attributes.get("help", "")

class YamlAction(argparse.Action):
  """Yaml Action
  
  A custom action for argparse that loads a yaml file into a python object. 
  This is intended to be used alongside the File type in argparse. This way
  the file type (yaml) is enforced. 
  """
  def __init__(self, option_strings, dest, nargs=None, **kwargs):
    super().__init__(option_strings, dest, **kwargs)
  def __call__(self, parser, namespace, value, option_string=None):
    data = yaml.load(value, Loader=yaml.FullLoader)
    setattr(namespace, self.dest, data)

class JsonPatchAction(argparse._AppendAction):
  """Json Patch Action

  A custom argparse action that translates [JSON Patch](https://jsonpatch.com/) operations from the command line arguments. This allows users to specify operations like `add`, `remove`, `copy`, `replace`, `test`, and `move` directly in the command line.

  To add a key/value pair, use `--add <key> <value>`.
  To remove a key, use `--remove <key>`.
  To copy a key from one path to another, use `--copy <from> <to>`.
  To replace a key's value, use `--replace <key> <value>`.
  To test a key's value, use `--test <key> <value>`.
  To move a key from one path to another, use `--move <from> <to>`.
  """
  def __init__(self, option_strings, dest, nargs='+', metavar=('key', 'value'), **kwargs):
    opts = ["--add", "--remove", "--copy", "--replace", "--test", "--move"]
    super().__init__(opts, dest, nargs=nargs, metavar=metavar, **kwargs)
  def __call__(self, parser, namespace, value, option_string=None):
    def validate_key(key):
      key = key.replace("[", "/").replace("]", "")
      key = "/" + key if key[0] != "/" else key
      return key
    def validate_value(v):
      try:
        return json.loads(v)
      except json.JSONDecodeError:
        try: # attempt to load again with literal double quotes
          return json.loads('"' + v + '"')
        except json.JSONDecodeError: # still not so error
          raise DuploError(f"Invalid JSON value for {op} operation.")
    patch = None
    key = validate_key(value[0])
    op = option_string[2:]
    if op in ["add", "replace", "test"]:
      patch = {"op": op, "path": key, "value": validate_value(value[1])}
    elif op in ["remove"]:
      patch = {"op": op, "path": key}
    elif op in ["copy", "move"]:
      patch = {"op": op, "from": key, "path": validate_key(value[1])}
    super().__call__(parser, namespace, patch, option_string)

class DataMapAction(argparse.Action):
  """Data Map Action
  
  A custom argparse action which allows for making key/value pairs from either literal values or files.
  This is fashioned after the kubectl params for secrets and configmaps, where you can specify a file or a literal value.

  For files you specify `--from-file <file name>` which will use the file name as the key and the contents of the file as the value. If you want to change the name of the key, then simply specify `--from-file <key>=<file name>`. If you want to use stdin, then use `--from-file -` which will use "stdin" as the key. You can even do `--from-file foo.txt=-` to use stdin and name the key.

  For literal values you specify `--from-literal <key>=<value>`. This will create a key/value pair in the data map. In this case the key and value must always be specified. 
  """
  def __init__(self, option_strings, dest, nargs='+', **kwargs):
    super(DataMapAction, self).__init__(option_strings, dest, nargs=nargs, **kwargs)
    self.__filetype = argparse.FileType()
  def __call__(self, parser, namespace, values, option_string=None):
    key = None 
    value = None
    items = getattr(namespace, self.dest, None)
    data = items if items else {}
    if option_string == "--from-file":
      key, value = self.__file_value(values[0])
    elif option_string == "--from-literal":
      key, value = self.__literal_value(values[0])
    data[key] = value
    setattr(namespace, self.dest, data)

  def __file_value(self, string):
    key = None
    fpath = None
    parts = string.split("=", 1)
    if len(parts) == 1:
      fpath = parts[0]
      key = "stdin" if fpath == "-" else os.path.basename(fpath)
    elif len(parts) == 2:
      key, fpath = parts
    f = self.__filetype(fpath)
    value = f.read().strip()
    return [key, value]
  
  def __literal_value(self, string):
    parts = string.split("=", 1)
    if len(parts) != 2:
      raise argparse.ArgumentTypeError("Literal values must be in the format key=value")
    return parts



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Resource


@Resource("pvc")
class PersistentVolumeClaim(DuploTenantResourceV3):
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "k8s/pvc")

  def name_from_body(self, body):
    return body["name"]


import pytest
from typing import Any, Dict, Optional, Callable
from duplocloud.errors import DuploError

def execute_test(func: Callable, *args, **kwargs) -> Any:
    """Standardized test execution with error handling.
    
    Args:
        func: Function to execute
        *args: Positional arguments for the function
        **kwargs: Keyword arguments for the function
        
    Returns:
        Any: Result of the function execution
        
    Raises:
        pytest.fail: If the function execution fails
    """
    try:
        return func(*args, **kwargs)
    except DuploError as e:
        pytest.fail(f"Test failed: {e}")

def assert_response(response: Dict[str, Any], 
                   expected_message: str,
                   status_code: int = 200,
                   exact_match: bool = True) -> None:
    """Standard response validation.
    
    Args:
        response: Response dictionary to validate
        expected_message: Expected message in the response
        status_code: Expected status code (default: 200)
        exact_match: Whether to check for exact message match (default: True)
    """
    assert response is not None, "Response should not be None"
    assert "message" in response, "Response should contain 'message'"
    
    if exact_match:
        assert response["message"] == expected_message
    else:
        assert expected_message in response["message"]
        
    if "status_code" in response:
        assert response["status_code"] == status_code

def get_resource_name(tenant_name: str, 
                     resource_type: str, 
                     suffix: Optional[str] = None) -> str:
    """Generate standard resource name.
    
    Args:
        tenant_name: Name of the tenant
        resource_type: Type of resource
        suffix: Optional suffix for the resource name
        
    Returns:
        str: Generated resource name
    """
    name = f"duploservices-{tenant_name}-{resource_type}"
    if suffix:
        name = f"{name}-{suffix}"
    return name

def setup_resource_fixture(duplo: Any, 
                         resource_type: str,
                         request: Any,
                         suffix: Optional[str] = None) -> tuple:
    """Standard resource fixture setup.
    
    Args:
        duplo: DuploClient instance
        resource_type: Type of resource to load
        request: pytest request object
        suffix: Optional suffix for resource name
        
    Returns:
        tuple: (resource, resource_name)
    """
    resource = duplo.load(resource_type)
    tenant_name = resource.tenant["AccountName"]
    resource_name = get_resource_name(tenant_name, resource_type, suffix)
    if request and hasattr(request, "cls"):
        request.cls.resource_name = resource_name
    return resource, resource_name


from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.errors import DuploError
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("batch_definition")
class DuploBatchDefinition(DuploTenantResourceV3):
  """Manage AWS Batch Job Definition Resources

  Manage batch Job Definitions as a resource in Duplo.

  Read more docs here:
  https://docs.duplocloud.com/docs/overview/aws-services/batch
  """

  def __init__(self, duplo: DuploClient):
    super().__init__(duplo,
                     slug="aws/batchJobDefinition",
                     prefixed=True)

  def name_from_body(self, body):
    return body["JobDefinitionName"]

  @Command()
  def list(self) -> list:
    """List all Batch Job Definitions.

    Usage:
      ```sh
      duploctl batch_definition list
      ```

    Example:
      Retrieve all of the revision IDs for one job definition using JMESPATH query and output as yaml
      ```sh
      duploctl batch_definition list --query "[?JobDefinitionName=='duploservices-dev01-myjobdef'].Revision" -o yaml
      ```

    Returns:
      list: A list of Batch Job Definitions.
    """
    return super().list()

  @Command()
  def find(self,
           name: args.NAME,
           to_revision: args.TO_REVISION = None) -> dict:
    """Find a Single Batch Job Definition by name.

    Usage: cli usage
      ```sh
      duploctl batch_definition find <name>
      ```

    Example:
      Find the previous revision to the latest. This is what you would use to do a rollback.
      ```sh
      duploctl batch_definition find myjobdef --revision -2
      ```

    Example:
      Find the earliest revision of a job definition.
      ```sh
      duploctl batch_definition find myjobdef --revision 0
      ```

    Args:
      name: The name of the Batch Job Definition to find.
      to_revision: The specific revision of the Batch Job Definition to find. If negative it will walk back that number of revisions from whatever number is the highest revision. The default is -1 which will return the latest.

    Returns:
      resource: The Batch Job Definition object.
    """
    # latest will be the default revision, which is -1
    to_rid = -1 if to_revision is None else to_revision
    rid = None
    definitions = self.list()
    n = self.prefixed_name(name)
    revisions = {
      d["Revision"]: d for d in definitions if self.name_from_body(d) == n
    }
    rids = sorted(revisions.keys())
    if len(rids) == 0:
      raise DuploError(f"Batch Job Definition '{name}' not found", 404)

    # if to_revision is negative or 0, we will walk back that many revisions
    if to_rid <= 0:
      # try but catch the index out of range exception
      try:
        rid = rids[to_rid]
      except IndexError:
        raise DuploError(f"Batch Job Definition '{name}' does not go back {to_rid} revisions. The furthest back is -{len(rids)} to revision number {rids[0]}", 404)
    # if to_revision is positive, we will just find that revision if it exists in the list
    elif to_rid > 0:
      rid = to_rid

    # finally return it if it exists
    if rid not in rids:
      raise DuploError(f"Batch Job Definition '{name}' revision {rid} not found. The latest is {rids[-1]}", 404)
    else:
      return revisions[rid]

  @Command()
  def create(self,
             body: args.BODY) -> dict:
    """Create a Batch Job Definition resource.

    Usage: CLI Usage
      ```sh
      duploctl batch_definition create -f 'batchdefinition.yaml'
      ```
      Contents of the `batchdefinition.yaml` file
      ```yaml
      --8<-- "src/tests/data/batchdefinition.yaml"
      ```

    Example: One liner example
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/batchdefinition.yaml"
      \"\"\" | duploctl batch_definition create -f -
      ```

    Args:
      body: The resource to create.

    Returns:
      message: Success message.

    Raises:
      DuploError: If the resource could not be created.
    """
    arn = super().create(body)
    rid = int(arn.split(":")[-1])
    name = self.name_from_body(body)
    return {
      "Message": f"Batch Job Definition '{name}' created successfully.",
      "JobDefinitionArn": arn,
      "Revision": rid
    }

  @Command()
  def delete(self,
              name: args.NAME,
              to_revision: args.TO_REVISION = None,
              all: args.ALL = False) -> dict:
      """Delete a Batch Job Definition by name.

      Usage: CLI Usage
        ```sh
        duploctl batch_definition delete <name>
        ```

      Args:
        name: The name of the Batch Job Definition to delete.
        to_revision: The specific revision of the Batch Job Definition to delete. If negative it will walk back that number of revisions from whatever number is the highest revision.

      Returns:
        message: Success message.
      """
      n = self.prefixed_name(name)
      if all:
        definitions = self.list()
        rids = [
          d["Revision"] for d in definitions if self.name_from_body(d) == n
        ]
        for rid in rids:
          super().delete(f"{n}:{rid}")
        msg = f"Batch Job Definition '{name}' deleted successfully along with the following revisions: {', '.join(rids)}."
      else:
        resource = self.find(n, to_revision)
        rid = resource["Revision"]
        super().delete(f"{n}:{rid}")
        msg = f"Batch Job Definition '{name}' deleted revision {rid} successfully."
      return {
        "Message": msg
      }

  @Command()
  def update_image(self,
                  name: args.NAME,
                  image: args.IMAGE) -> dict:
    """Update the image of a Batch Job Definition by name.

    Usage: CLI Usage
      ```sh
      duploctl batch_definition update_image <name> --image <image>
      ```

    Args:
      name: The name of the Batch Job Definition to update.
      image: The new image to set for the Batch Job Definition.

    Returns:
      message: Success message.
    """
    resource = self._to_def_request(self.find(name))
    resource["ContainerProperties"]["Image"] = image
    res = self.create(resource)
    res["Message"] = f"Batch Job Definition '{name}' updated successfully to revision {res['Revision']} with new image '{image}'."
    return res

  def _to_def_request(self, body: dict) -> dict:
    # Remove top level fields if present
    body.pop("ContainerOrchestrationType", None)
    body.pop("JobDefinitionArn", None)
    body.pop("Revision", None)
    body.pop("Status", None)
    if body.get("SchedulingPriority") == 0:
      body.pop("SchedulingPriority", None)
    # Remove container fields
    if "ContainerProperties" in body:
        container_props = body["ContainerProperties"]
        container_props.pop("JobRoleArn", None)
        # If VCPU and MEMORY are set in res reqs, remove them from the container props
        resource_reqs = container_props.get("ResourceRequirements", [])
        if any(self._type_matches(req, "VCPU") for req in resource_reqs):
            container_props.pop("Vcpus", None)
        if any(self._type_matches(req, "MEMORY") for req in resource_reqs):
            container_props.pop("Memory", None)
    return body

# DuploCloud returns ResourceRequirements in an incorrect format:
# "ResourceRequirements": [
#     {
#         "Type": {
#             "Value": "MEMORY"
#         },
#         "Value": "2048"
#     },
#     {
#         "Type": {
#             "Value": "VCPU"
#         },
#         "Value": "1"
#     }
# ]
# This check handles both the correct structure eg, "Type": "MEMORY"
# and the malformed structure, eg ("Type": {"Value": "MEMORY"}).
  def _type_matches(self, req, value):
    t = req.get("Type")
    if isinstance(t, dict):
        t = t.get("Value")
    return t == value



from git import Repo
from packaging.version import Version
import semver
import os
import datetime
from jinja2 import Template

CHANGELOG = 'CHANGELOG.md'
REPO_URL = "https://github.com/duplocloud/duploctl"

class Project:
  def __init__(self):
    self.__latest_tag = None
    self.__changelog = None
    self.cwd = os.getcwd()
    self.repo = Repo(self.cwd)

  @property
  def ref(self):
    return self.repo.head.reference
  
  @property
  def changelog(self):
    if not self.__changelog:
      with open(CHANGELOG, 'r') as f:
        self.__changelog = f.read()
    return self.__changelog
  
  @property
  def latest_tag(self):
    if not self.__latest_tag:
      latest = Version("0.0.0")
      for t in self.repo.tags:
        v = Version(t.name[1:])
        if v > latest:
          latest = v
      self.__latest_tag = semver.VersionInfo.parse(str(latest))
    return self.__latest_tag
  
  def bump_version(self, action: str="patch"):
    latest = self.latest_tag
    if action == "major":
      return latest.bump_major()
    if action == "minor":
      return latest.bump_minor()
    if action == "patch":
      return latest.bump_patch()
    return latest
  
  def dist_file(self, file, content):
    os.makedirs('dist', exist_ok=True)
    with open(f"dist/{file}", 'w') as f:
      f.write(content)

  def release_notes(self, version="Unreleased"):
    header = f"## [{version}]"
    inblock = False
    msg = []
    for line in self.changelog.splitlines():
      if line.startswith(header):
        inblock = True
        continue
      if inblock and line.startswith("## ["):
        break
      if inblock:
        msg.append(line.strip())
    return "\n".join(msg)
  
  def install_notes(self, version):
    # get the install.md file within the wiki folder
    with open("wiki/Installation.md") as f:
      tpl = Template(f.read())
      return tpl.render(version=version, repo_url=REPO_URL)
  
  def reset_changelog(self, version):
    unreleased = "## [Unreleased]"
    # get todays date formatted like 2023-03-05
    date = datetime.datetime.now().strftime("%Y-%m-%d")
    new_notes = f"{unreleased}\n\n## [{version}] - {date}"
    return self.changelog.replace(unreleased, new_notes)



import pytest
from duplocloud.errors import DuploError
from .conftest import get_test_data

@pytest.fixture(scope="class")
def configmap_resource(duplo, request):
    """Fixture to load the ConfigMap resource and ensure ConfigMap name persists across tests."""
    resource = duplo.load("configmap")
    request.cls.configmap_name = f"duploctl"
    return resource

def execute_test(func, *args, **kwargs):
    """Helper function to execute a test and handle errors."""
    try:
        return func(*args, **kwargs)
    except DuploError as e:
        pytest.fail(f"Test failed: {e}")

@pytest.mark.usefixtures("configmap_resource")
class TestConfigMap:

    @pytest.mark.integration
    @pytest.mark.dependency(name="create_configmap", scope="class")
    @pytest.mark.order(5)
    def test_create_configmap(self, configmap_resource):
        """Test creating a ConfigMap with various methods."""
        body = get_test_data("configmap")
        response = execute_test(configmap_resource.create, name=self.configmap_name, body=body)
        assert response["metadata"]["name"] == self.configmap_name
        assert response["data"] == body["data"]
        # Test creating with complete body
        body = {
            "metadata": {"name": f"{self.configmap_name}-2"},
            "data": {"config1": "test1", "config2": "test2"}
        }
        response = execute_test(configmap_resource.create, body=body)
        assert response["metadata"]["name"] == f"{self.configmap_name}-2"
        assert response["data"] == body["data"]
        # Test dryrun option
        dryrun_response = execute_test(configmap_resource.create, name=f"{self.configmap_name}-3", 
                                     data={"test": "value"}, dryrun=True)
        assert dryrun_response["metadata"]["name"] == f"{self.configmap_name}-3"
        assert "data" in dryrun_response

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_configmap"], scope="class")
    @pytest.mark.order(6)
    def test_find_configmap(self, configmap_resource):
        """Test finding a ConfigMap."""
        # Test finding existing ConfigMap
        configmap = execute_test(configmap_resource.find, self.configmap_name)
        assert configmap["metadata"]["name"] == self.configmap_name
        assert "data" in configmap
        assert configmap["data"]["foo"] == "bar"

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_configmap"], scope="class")
    @pytest.mark.order(7)
    def test_update_configmap(self, configmap_resource):
        """Test updating a ConfigMap using different methods."""
        # Test updating with new data
        new_data = {"key3": "value3", "key1": "updated1"}
        response = execute_test(configmap_resource.update, name=self.configmap_name, data=new_data)
        assert response["data"]["key3"] == "value3"
        assert response["data"]["key1"] == "updated1"
        assert response["data"]["foo"] == "bar"
        # Test updating with patches
        patches = [
            {'op': 'add', 'path': '/data/new_key', 'value': 'new_value'},
            {'op': 'remove', 'path': '/data/foo', 'value': None},
            {'op': 'replace', 'path': '/data/key1', 'value': 'final_value'}
        ]
        response = execute_test(configmap_resource.update, name=self.configmap_name, patches=patches)
        assert response["data"]["new_key"] == "new_value"
        assert "key2" not in response["data"]
        assert response["data"]["key1"] == "final_value"
        # Test dryrun update
        dryrun_response = execute_test(configmap_resource.update, name=self.configmap_name,
                                     data={"test": "value"}, dryrun=True)
        assert "test" in dryrun_response["data"]

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_configmap"], scope="class")
    @pytest.mark.order(8)
    def test_delete_configmap(self, configmap_resource):
        """Test deleting ConfigMaps and verifying deletion."""
        # Delete first ConfigMap
        response = execute_test(configmap_resource.delete, name=self.configmap_name)
        assert response["message"] == f"Successfully deleted configmap '{self.configmap_name}'"
        # Delete second ConfigMap
        response = execute_test(configmap_resource.delete, name=f"{self.configmap_name}-2")
        assert response["message"] == f"Successfully deleted configmap '{self.configmap_name}-2'"



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.errors import DuploError, DuploFailedResource, DuploStillWaiting
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("cloudfront")
class DuploCloudFront(DuploTenantResourceV3):
  """Manage CloudFront Distributions

  Configuring a CloudFront distributions in DuploCloud are content delivery network (CDN) configurations that
  help you improve the performance of your web applications.

  See more details at: https://docs.duplocloud.com/docs/overview/aws-services/cloudfront
  """

  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "aws/cloudFrontDistribution")

  def wait_check(self, distribution_id):
    status_response = self.find(distribution_id)
    status = status_response.get("Distribution", {}).get("Status", "Unknown").lower()
    if status == "deployed":
      return status_response
    elif status in ["failed", "error"]:
      raise DuploFailedResource(f"CloudFront distribution {distribution_id} failed to deploy.")
    else:
      raise DuploStillWaiting(f"Timed out waiting for CloudFront {distribution_id} to be deployed.")

  @Command()
  def find(self, distribution_id: args.DISTRIBUTION_ID):
    """Find a CloudFront distribution.

    Find a CloudFront distribution by its distribution ID.

    Usage:
      ```sh
      duploctl cloudfront find <distribution_id>
      ```

    Args:
      distribution_id: The CloudFront distribution ID.

    Returns:
      dict: The service object.
    """
    response = self.duplo.get(self.endpoint(name=distribution_id))
    response.raise_for_status()
    return response.json()

  @Command()
  def create(self, body: args.BODY):
    """Create a CloudFront distribution.

    Creates a new CloudFront distribution with the specified configuration.

    Usage: Basic CLI Use
      ```sh
      duploctl cloudfront create -f cloudfront.yaml
      ```

    Example cloudfront.yaml:
      ```yaml
      --8<-- "src/tests/data/cloudfront-create.yaml"
      ```

    Args:
      body: The distribution configuration including origins, behaviors, and other settings.

    Returns:
      dict: The created distribution details including the distribution ID.

    Raises:
      DuploError: If the distribution could not be created or the configuration is invalid.
      DuploFailedResource: If the distribution creation process fails.
    """
    data = None
    try:
      response = self.duplo.post(self.endpoint(), body)
      response.raise_for_status()
      data = response.json()
      distribution_id = data.get("Id") or data.get("Distribution", {}).get("Id")
      if not distribution_id:
        raise ValueError("Failed to retrieve CloudFront Distribution ID from response.")
      if self.duplo.wait:
        self.wait(lambda: self.wait_check(distribution_id) is not None, 1200)
      return data
    except Exception as e:
      raise DuploError(f"Failed to create CloudFront distribution: {e}")

  @Command()
  def update(self, body: args.BODY):
    """Update a CloudFront distribution.

    Updates an existing CloudFront distribution with new configuration settings.

    Usage: Basic CLI Use
      ```sh
      duploctl cloudfront update -f cloudfront.yaml
      ```

    Example cloudfront.yaml:
      ```yaml
      --8<-- "src/tests/data/cloudfront-update.yaml"
      ```

    Args:
      body: The updated distribution configuration.

    Returns:
      dict: The updated distribution details.

    Raises:
      DuploError: If the distribution could not be updated or the configuration is invalid.
      DuploFailedResource: If the update process fails.
    """
    try:
      response = self.duplo.put(self.endpoint(), body)
      response.raise_for_status()
      data = response.json()
      distribution_id = data.get("Id") or data.get("Distribution", {}).get("Id")
      if not distribution_id:
        raise ValueError("Failed to retrieve CloudFront Distribution ID from response.")
      if self.duplo.wait:
        self.wait(lambda: self.wait_check(distribution_id) is not None, 1200)
      return data
    except Exception as e:
      raise DuploError(f"Failed to update CloudFront distribution: {e}")

  @Command()
  def disable(self, distribution_id: args.DISTRIBUTION_ID):
    """Disable a CloudFront distribution.

    Disables content delivery for a CloudFront distribution. When disabled,
    the distribution will stop serving content but will retain its configuration.

    Usage: Basic CLI Use
      ```sh
      duploctl cloudfront disable <distribution-id>
      ```

    Args:
      distribution_id: The ID of the CloudFront distribution to disable.

    Returns:
      dict: The updated distribution details.

    Raises:
      DuploError: If the distribution could not be disabled.
      DuploFailedResource: If the disable process fails.
    """
    body = {"Id":distribution_id,"DistributionConfig":{"Disabled":"true"}}
    response = self.duplo.put(self.endpoint(), body)
    if self.duplo.wait:
      self.wait(lambda: self.wait_check(distribution_id) is not None, 1200)
    return response.json()
  
  @Command()
  def enable(self, distribution_id: args.DISTRIBUTION_ID):
    """Enable a CloudFront distribution.

    Enables content delivery for a previously disabled CloudFront distribution.
    When enabled, the distribution will start serving content using its current
    configuration.

    Usage: Basic CLI Use
      ```sh
      duploctl cloudfront enable <distribution-id>
      ```

    Args:
      distribution_id: The ID of the CloudFront distribution to enable.

    Returns:
      dict: The updated distribution details.

    Raises:
      DuploError: If the distribution could not be enabled.
      DuploFailedResource: If the enable process fails.
    """
    body = {"Id":distribution_id,"DistributionConfig":{"Enabled":"true"}}
    response = self.duplo.put(self.endpoint(), body)
    if self.duplo.wait:
      self.wait(lambda: self.wait_check(distribution_id) is not None, 1200)
    return response.json()

  @Command()
  def delete(self, distribution_id: args.DISTRIBUTION_ID):
    """Delete a CloudFront distribution.

    Permanently removes a CloudFront distribution. This operation is irreversible
    and will remove all distribution settings and cached content. The distribution
    must be disabled before it can be deleted.

    Usage: Basic CLI Use
      ```sh
      duploctl cloudfront delete <distribution-id>
      ```

    Args:
      distribution_id: The ID of the CloudFront distribution to delete.

    Returns:
      dict: Success message confirming the deletion.

    Raises:
      DuploError: If the distribution could not be deleted or is not in a deletable state.
    """
    response = self.duplo.delete(self.endpoint(distribution_id))
    response.raise_for_status()
    return {"message": f"CloudFront distribution {distribution_id} deleted"}

  @Command()
  def get_status(self, distribution_id: args.DISTRIBUTION_ID):
    """Get the current status of a CloudFront distribution.

    Retrieves the current deployment status of a CloudFront distribution.
    The status indicates whether the distribution is deployed, in progress,
    or in an error state.

    Usage: Basic CLI Use
      ```sh
      duploctl cloudfront get-status <distribution-id>
      ```

    Args:
      distribution_id: The ID of the CloudFront distribution to check.

    Returns:
      str: The current status of the distribution (e.g., 'Deployed', 'InProgress').

    Raises:
      DuploError: If the distribution could not be found or the status is unavailable.
    """
    response = self.find(distribution_id)
    status = response.get("Distribution", {}).get("Status")
    if status is None:
      raise DuploError(f"Status not found for CloudFront distribution {distribution_id}")
    return status



import pytest
from duplocloud.errors import DuploError
from .conftest import get_test_data

@pytest.fixture(scope="class")
def ingress_resource(duplo, request):
    """Fixture to load an Ingress resource."""
    resource = duplo.load("ingress")
    request.cls.ingress_name = None
    return resource

def execute_test(func, *args, **kwargs):
    """Helper function to execute a test and handle errors."""
    try:
        return func(*args, **kwargs)
    except DuploError as e:
        pytest.fail(f"Test failed: {e}")

class TestIngress:

    @pytest.mark.integration
    @pytest.mark.dependency(name="create_ingress", scope="session")
    @pytest.mark.order(6)
    def test_create_ingress(self, ingress_resource, request):
        """Test creating an Ingress resource."""
        r = ingress_resource
        body = get_test_data("ingress")
        response = execute_test(r.create, body=body)
        assert response.get("message") and f"Successfully Created an Ingress '{body['name']}'" in response["message"], "Ingress creation failed"
        request.cls.ingress_name = body['name']

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_ingress"], scope="session")
    @pytest.mark.order(7)
    def test_update_ingress(self, ingress_resource):
        """Test updating an Ingress resource."""
        r = ingress_resource
        update_body = get_test_data("ingress")
        response = execute_test(r.update, name=self.ingress_name, body=update_body)
        assert response.get("message") and f"Successfully Updated an Ingress '{update_body['name']}'" in response["message"], "Ingress updation failed"

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_ingress"], scope="session")
    @pytest.mark.order(8)
    def test_list_ingress(self, ingress_resource):
        """Test listing an Ingress."""
        r = ingress_resource
        ingresses = execute_test(r.list)
        assert isinstance(ingresses, list), "Ingress list response is not a list"
        assert len(ingresses) > 0, "No Ingress found"

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_ingress"], scope="session")
    @pytest.mark.order(997)
    def test_delete_ingress(self, ingress_resource):
        """Test deleting an Ingress."""
        r = ingress_resource
        assert self.ingress_name is not None, "Ingress name not found!"
        response = execute_test(r.delete, name=self.ingress_name)
        assert response.get("message") == f"k8s/ingress/{self.ingress_name} deleted"



import pytest
import time
from duplocloud.errors import DuploError
from .conftest import get_test_data

@pytest.fixture(scope="class")
def cloudfront_resource(duplo, request):
    """Fixture to load the CloudFront resource and ensure CloudFront ID persists across tests."""
    resource = duplo.load("cloudfront")
    resource.duplo.wait = True
    tenant = resource.tenant["AccountName"]
    request.cls.cloudfront_id = None
    return resource

def execute_test(func, *args, **kwargs):
    """Helper function to execute a test and handle errors."""
    try:
        return func(*args, **kwargs)
    except DuploError as e:
        pytest.fail(f"Test failed: {e}")

@pytest.mark.usefixtures("cloudfront_resource")
class TestCloudFront:

    @pytest.mark.integration
    @pytest.mark.dependency(name="create_cloudfront", scope="class")
    @pytest.mark.order(5)
    def test_create_cloudfront(self, cloudfront_resource, request):
        """Test creating a CloudFront distribution and store ID at class level."""
        r = cloudfront_resource
        body = get_test_data("cloudfront-create")
        response = execute_test(r.create, body=body)
        status_response = execute_test(r.get_status, distribution_id=response["Id"])
        assert status_response == "Deployed", "CloudFront distribution not deployed"
        request.cls.cloudfront_id = response["Id"]
        print(f"CloudFront Created! ID: {request.cls.cloudfront_id}")

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_cloudfront"], scope="class")
    @pytest.mark.order(6)
    def test_update_cloudfront(self, cloudfront_resource, request):
        """Test updating a CloudFront distribution using stored CloudFront ID."""
        r = cloudfront_resource
        assert self.cloudfront_id is not None, "CloudFront ID not found! Ensure test_create_cloudfront ran successfully."
        update_body = get_test_data("cloudfront-update")
        cloudfront = execute_test(r.find, distribution_id=self.cloudfront_id)
        update_body["Id"] = self.cloudfront_id
        update_body["DistributionConfig"]["CallerReference"] = cloudfront["Distribution"]["DistributionConfig"]["CallerReference"]
        response = execute_test(r.update, body=update_body)
        status_response = execute_test(r.get_status, distribution_id=response["Id"])
        assert status_response == "Deployed", "CloudFront distribution not deployed after update"

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_cloudfront"], scope="session")
    @pytest.mark.order(7)
    def test_list_cloudfront(self, cloudfront_resource):
        """Test listing CloudFront distributions."""
        r = cloudfront_resource
        distributions = execute_test(r.list)
        assert isinstance(distributions, list), "CloudFront list response is not a list"
        assert len(distributions) > 0, "No CloudFront distributions found"

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_cloudfront"], scope="session")
    @pytest.mark.order(8)
    def test_disable_cloudfront(self, cloudfront_resource):
        """Test disabling a CloudFront distribution."""
        r = cloudfront_resource
        assert self.cloudfront_id is not None, "CloudFront ID not found!"
        execute_test(r.disable, distribution_id=self.cloudfront_id)
        status_response = execute_test(r.get_status, distribution_id=self.cloudfront_id)
        assert status_response == "Deployed", "CloudFront distribution was not disabled"

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_cloudfront"], scope="session")
    @pytest.mark.order(9)
    def test_enable_cloudfront(self, cloudfront_resource):
        """Test enabling a CloudFront distribution."""
        r = cloudfront_resource
        assert self.cloudfront_id is not None, "CloudFront ID not found!"
        execute_test(r.enable, distribution_id=self.cloudfront_id)
        status_response = execute_test(r.get_status, distribution_id=self.cloudfront_id)
        assert status_response == "Deployed", "CloudFront distribution was not enabled"
        execute_test(r.disable, distribution_id=self.cloudfront_id)

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_cloudfront"], scope="session")
    @pytest.mark.order(997)
    def test_delete_cloudfront(self, cloudfront_resource):
        """Test deleting a CloudFront distribution."""
        r = cloudfront_resource
        assert self.cloudfront_id is not None, "CloudFront ID not found!"
        response = execute_test(r.delete, distribution_id=self.cloudfront_id)
        assert response.get("message") == f"CloudFront distribution {self.cloudfront_id} deleted"



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.errors import DuploError
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("batch_job")
class DuploBatchJob(DuploTenantResourceV3):
  """Manage AWS Batch Job Resources

  Run batch jobs as a managed service on AWS infrastructure. 

  Read more docs here: 
  https://docs.duplocloud.com/docs/overview/aws-services/batch
  """

  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, 
                     slug="aws/batchJobs",
                     prefixed=True)

  def name_from_body(self, body):
    return body["JobName"]

  @Command()
  def list(self,
           queue: args.BATCH_QUEUE = None,
           all: args.ALL = False) -> list:
    """List all Batch Job Definitions.

    Usage:
      ```sh
      duploctl batch_job list <queue_name>
      ```
    
    Args:
      queue_name: The name of the Batch Queue for this list of jobs.
      all: If true, list all jobs in all queues.

    Returns:
      list: A list of Batch Job Definitions.
    """
    def get_job_list(qn):
      response = self.duplo.get(self.endpoint(qn))
      return response.json()
    if not queue and not all:
      raise DuploError("You must specify a queue name with --queue <queue_name> or use --all to list jobs in all queues.")
    if queue:
      return get_job_list(self.prefixed_name(queue))
    else:
      queue_svc = self.duplo.load("batch_queue")
      queues = queue_svc.list()
      jobs = []
      for q in queues:
        j = get_job_list(queue_svc.name_from_body(q))
        jobs.extend(j)
      return jobs

  @Command()
  def find(self, 
           name: args.NAME = None,
           queue: args.BATCH_QUEUE = None) -> dict:
    """Find a Single Batch Job by name.

    Usage: cli usage
      ```sh
      duploctl batch_definition find <name> <queue_name>
      ```

    Args:
      name: The name of the Batch Job Definition to find.
      queue_name: The name of the Batch Queue for this job.
      to_revision: The specific revision of the Batch Job Definition to find. If negative it will walk back that number of revisions from whatever number is the highest revision.

    Returns: 
      resource: The Batch Job Definition object.
    """
    jobs = self.list(queue)
    n = self.prefixed_name(name)
    for job in jobs:
      if self.name_from_body(job) == n:
        return job
    raise DuploError(f"Batch Job '{name}' not found in queue '{queue}'", 404)

  @Command()
  def create(self, 
             body: args.BODY) -> dict:
    """Create a Batch Job resource.

    Usage: CLI Usage
      ```sh
      duploctl batch_job create -f 'batchjob.yaml'
      ```
      Contents of the `batchjob.yaml` file
      ```yaml
      --8<-- "src/tests/data/batchjob.yaml"
      ```

    Example: One liner example
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/batchjob.yaml"
      \"\"\" | duploctl batch_job create -f -
      ```
    
    Args:
      body: The resource to create.

    Returns: 
      message: Success message.

    Raises:
      DuploError: If the resource could not be created.
    """
    queue = body.get("JobQueue", None)
    definition = body.get("JobDefinition", None)
    if not (queue and definition):
      raise DuploError("You must specify both JobQueue and JobDefinition in the body to create a Batch Job.")
    body["JobQueue"] = self.prefixed_name(queue)
    if not definition.startswith("arn:aws:batch:"):
      body["JobDefinition"] = self._job_definition_arn(definition)
    id = super().create(body)
    return {
      "Message": f"Batch Job '{self.name_from_body(body)}' created successfully.",
      "JobId": id
    }

  def _job_definition_arn(self, definition: str) -> str:
    def_name_parts = definition.split(":")
    # make sure there is either one or two parts only
    if len(def_name_parts) not in [1, 2]:
      raise DuploError(f"When the JobDefinition is not an arn, the name may only have one optional colon where the value after is the revision id: {definition}")
    def_name = self.prefixed_name(def_name_parts[0])
    rid = def_name_parts[1] if len(def_name_parts) == 2 else None
    # find the def using the def service
    job_def_svc = self.duplo.load("batch_definition")
    job_def = job_def_svc.find(def_name, rid)
    return job_def["JobDefinitionArn"]



from duplocloud.client import DuploClient
from duplocloud.errors import DuploError, DuploExpiredCache
from duplocloud.resource import DuploResource
from duplocloud.commander import Command, Resource
import duplocloud.args as args
import os
from pathlib import Path
import yaml
import configparser
import webbrowser
from datetime import datetime
import jwt

INSTALL_HINT = """
Install duploctl for use with kubectl by following
https://cli.duplocloud.com/Jit/
"""

@Resource("jit")
class DuploJit(DuploResource):
  """Just In Time (JIT) Resource

  Just in time access for AWS. This will use Duplo credentials to ask a certain Duplo portal for temporary AWS credentials. These credentials will be valid for a certain amount of time and will be used to access AWS resources.

  Usage:
    using the `duploctl` command line tool, you can manage services with actions:

    ```sh
    duploctl jit <action>
    ```
  """
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)

  @Command()
  def token(self) -> dict:
    """Get JWT Token

    Get the JWT token for the current user. This is the token that is used to authenticate with the Duplo API.

    Usage:
      ```sh
      duploctl jit token
      ```

    Returns:
      token: The JWT token.
    """
    return {"token": self.duplo.token}

  @Command()
  def gcp(self, nocache: bool = None) -> dict:
    """GCP Access Token

    Get the GCP JWT token for the current user. This is the token that is used to authenticate with the GCP API. You must be an admin to use this feature.

    Example: Using for gcloud cli access
      Here is how to set the needed environment variables for the gcloud cli.

      ```sh
      for i in $(duploctl jit gcp -q '{CLOUDSDK_AUTH_ACCESS_TOKEN: Token, CLOUDSDK_CORE_PROJECT: ProjectId}' -o env); do export $i; done
      ```

    Usage:
      ```sh
      duploctl jit gcp
      ```

    Returns:
      token: The GCP JWT token.
    """
    k = self.duplo.cache_key_for("gcp-creds")
    nc = nocache if nocache is not None else self.duplo.nocache
    t = self.duplo.load("tenant")
    tenant = t.find()
    path = f"v3/admin/google/{tenant['TenantId']}/apiToken"
    # try and get those creds
    try:
      if nc:
        sts = self.duplo.get(path).json()
      else:
        sts = self.duplo.get_cached_item(k)
        if self.duplo.expired(sts.get("Expiration", None)):
          raise DuploExpiredCache(k)
    except DuploExpiredCache:
      sts = self.duplo.get(path).json()
      if "Expiration" not in sts:
        sts["Expiration"] = self.duplo.expiration()
      self.duplo.set_cached_item(k, sts)
    return sts

  @Command()
  def aws(self, nocache: bool = None) -> dict:
    """AWS STS Session Credentials

    Provides a full sts session with credentials and region. The default return is a valid exec credential for the AWS CLI. The global `--admin` flag can be used to get the credentials for an admin, or else per tenant
    credentials are returned. The `--interactive` flag can be used to get the credentials for an interactive session and use the cache.

    Basic Usage:
      ```sh
      duploctl jit aws
      ```

    Example: Using in AWS CLI Credential Process
      Here is an example for using the duploctl jit for aws in an AWS CLI config file.

      ```toml
      [profile myportal]
      region=us-west-2
      output=json
      credential_process=duploctl jit aws --host https://myportal.duplocloud.net --admin --interactive
      ```

    Example: Get AWS Environment Variables
      Here is an example using a query and env output to create some just in time aws credentials.

      ```sh
      duploctl jit aws -o env -q '{AWS_ACCESS_KEY_ID: AccessKeyId, AWS_SECRET_ACCESS_KEY: SecretAccessKey, AWS_SESSION_TOKEN: SessionToken, AWS_REGION: Region}'
      ```

      A one liner to export those credentials as environment variables.
      ```sh
      for i in `duploctl jit aws -q '{AWS_ACCESS_KEY_ID: AccessKeyId, AWS_SECRET_ACCESS_KEY: SecretAccessKey, AWS_SESSION_TOKEN: SessionToken, AWS_REGION: Region}' -o env`; do export $i; done
      ```

    Args:
      nocache (bool): Do not use cached credentials. Only for other methods to use.

    Returns:
      sts: The AWS STS session credentials.
    """
    sts = None
    path = None
    k = self.duplo.cache_key_for("aws-creds")
    nc = nocache if nocache is not None else self.duplo.nocache

    # check if admin or choose tenant
    if self.duplo.isadmin:
      path = "adminproxy/GetJITAwsConsoleAccessUrl"
    else:
      t = self.duplo.load("tenant")
      tenant = t.find()
      path = f"subscriptions/{tenant['TenantId']}/GetAwsConsoleTokenUrl"

    # try and get those creds
    try:
      if nc:
        sts = self.duplo.get(path).json()
      else:
        sts = self.duplo.get_cached_item(k)
        if self.duplo.expired(sts.get("Expiration", None)):
          raise DuploExpiredCache(k)
    except DuploExpiredCache:
      sts = self.duplo.get(path).json()
      if "Expiration" not in sts:
        sts["Expiration"] = self.duplo.expiration()
      self.duplo.set_cached_item(k, sts)
    sts["Version"] = 1
    return sts

  @Command()
  def update_aws_config(self,
                        name: args.NAME) -> dict:
    """Update AWS Config

    Update the AWS config file with a new profile. This will add a new profile to the AWS config file.
    This will honor the `AWS_CONFIG_FILE` environment variable if it is set.
    This will set the aws cli credentialprocess to use the `duploctl jit aws` command.
    The generated command will inherit the `--host`, `--admin`, and `--interactive` flags from the current command.

    Usage:
      ```sh
      duploctl jit update_aws_config myprofile
      ```

    Example: Add Admin Profile
      Run this command to add an admin profile.
      ```sh
      duploctl jit update_aws_config myportal --admin --interactive
      ```
      This generates the following in the AWS config file.
      ```toml
      [profile myportal]
      region = us-west-2
      credential_process = duploctl jit aws --host https://myportal.duplocloud.net --interactive --admin
      ```

    Args:
      name: The name of the profile to add.

    Returns:
      msg: The message that the profile was added.
    """
    config = os.environ.get("AWS_CONFIG_FILE", f"{Path.home()}/.aws/config")
    cp = configparser.ConfigParser()
    cp.read(config)
    name = name or "duplo"
    prf = f'profile {name}'
    cmd = self.duplo.build_command("duploctl", "jit", "aws")
    is_new_profile = not cp.has_section(prf)
    if is_new_profile:
      cp.add_section(prf)
    cp.set(prf, 'region', os.getenv("AWS_DEFAULT_REGION", "us-west-2"))
    cp.set(prf, 'credential_process', " ".join(cmd))
    os.makedirs(os.path.dirname(os.path.abspath(config)), exist_ok=True)
    with open(config, 'w') as configfile:
      cp.write(configfile)
    msg = f"aws profile named {name} was successfully {'added' if is_new_profile else 'updated'} in {config}"
    return {"message": msg}

  @Command()
  def web(self) -> dict:
    """Open Cloud Console

    Opens a default or specified browser to the Duploclouds underlying cloud.
    Currently this only supports AWS. The global `--browser` flag can be used to specify a browser.

    Usage:

    ```sh
    duploctl jit web --browser chrome
    ```

    Returns:
      msg: The message that the browser is opening.
    """
    b = self.duplo.browser
    wb = webbrowser if not b else webbrowser.get(b)
    sts = self.aws(nocache=True)
    wb.open(sts["ConsoleUrl"], new=0, autoraise=True)
    return {
      "message": "Opening AWS console in browser"
    }

  @Command()
  def k8s(self,
          planId: args.PLAN = None) -> dict:
    """Kubernetes JIT Exec Credentials

    Provides a full exec credential for kubectl. The default return is a valid exec credential for the kubectl CLI. The global `--admin` flag can be used to get the credentials for an admin, or else per tenant.
    An admin can pass the `--plan` or else it will be discovered from the chosen tenant. A non admin must
    choose a tenant.

    Usage:
      ```sh
      duploctl jit k8s
      ```

    Args:
      planId: The planId aka name the infrastructure.

    Returns:
      credentials: A Kubernetes client [ExecCredential](https://kubernetes.io/docs/reference/config-api/client-authentication.v1beta1/).
    """
    # either plan or tenant in cache key
    pt = planId or self.duplo.tenant or self.duplo.tenantid
    k = self.duplo.cache_key_for(f"plan,{pt},k8s-creds")
    creds = None
    try:
      if self.duplo.nocache:
        ctx = self.k8s_context(planId)
        creds = self.__k8s_exec_credential(ctx)
      else:
        creds = self.duplo.get_cached_item(k)
        exp = creds.get("status", {}).get("expirationTimestamp", None)
        if self.duplo.expired(exp):
          raise DuploExpiredCache(k)
    except DuploExpiredCache:
      ctx = self.k8s_context(planId)
      creds = self.__k8s_exec_credential(ctx)
      self.duplo.set_cached_item(k, creds)
    return creds

  @Command()
  def update_kubeconfig(self,
                        name: args.NAME = None,
                        planId: args.PLAN = None,
                        save: bool = True) -> dict:
    """Update Kubeconfig

    Update the kubeconfig file with a new context. This will add a new context to the kubeconfig file. This will honor the `KUBECONFIG` environment variable if it is set. The generated command will inherit the `--host`, `--admin`, and `--interactive` flags from the current command.

    Usage:
      ```sh
      duploctl jit update_kubeconfig --plan myplan
      ```

    Example: Add Admin Context
      Run this command to add an admin context.
      ```sh
      duploctl jit update_kubeconfig --plan myplan --admin --interactive
      ```
      This generates the following user credential process in the kubeconfig file.
      ```yaml
      users:
      - name: myplan
        user:
          exec:
            apiVersion: client.authentication.k8s.io/v1beta1
            args:
            - jit
            - k8s
            - --plan
            - myplan
            - --host
            - https://myportal.duplocloud.net
            - --admin
            - --interactive
            command: duploctl
            env: null
            installHint: |2

              Install duploctl for use with kubectl by following
              https://github.com/duplocloud/duploctl
            interactiveMode: IfAvailable
            provideClusterInfo: false
      ```

    Args:
      planId: The planId of the infrastructure.
      save: Save the kubeconfig file. This is a code only option.

    Returns:
      msg: The message that the kubeconfig was updated. Unless save is False, then the kubeconfig is returned.
    """
    # first get the kubeconfig file and parse it
    kubeconfig_path = os.environ.get("KUBECONFIG", f"{Path.home()}/.kube/config")
    kubeconfig = (yaml.safe_load(open(kubeconfig_path, "r"))
                  if os.path.exists(kubeconfig_path)
                  else self.__empty_kubeconfig())
    # load the cluster config info
    ctx = self.k8s_context(planId)
    ctx["PlanID"] = planId or ctx["Name"].removeprefix("duploinfra-")
    ctx["ARGS"] = ["jit", "k8s"]
    if self.duplo.isadmin:
      ctx["Name"] = name or ctx["PlanID"]
      ctx["ARGS"].extend(["--plan", ctx["PlanID"]])
      if self.duplo.tenant:
        ctx["DefaultNamespace"] = f"duploservices-{self.duplo.tenant}"
    else:
      ctx["Name"] = name or self.duplo.tenantid or self.duplo.tenant
    # add the cluster, user, and context to the kubeconfig
    self.__add_to_kubeconfig("clusters", self.__cluster_config(ctx), kubeconfig)
    self.__add_to_kubeconfig("users", self.__user_config(ctx), kubeconfig)
    self.__add_to_kubeconfig("contexts", self.__context_config(ctx), kubeconfig)
    kubeconfig["current-context"] = ctx["Name"]
    if save:
      os.makedirs(os.path.dirname(os.path.abspath(kubeconfig_path)), exist_ok=True)
      # write the kubeconfig back to the file
      with open(kubeconfig_path, "w") as f:
        yaml.safe_dump(kubeconfig, f)
      return {"message": f"kubeconfig updated successfully to {kubeconfig_path}"}
    else:
      return kubeconfig

  @Command()
  def k8s_context(self,
                  planId: args.PLAN = None) -> dict:
    """Get k8s context

    Gets context based on planId or tenant name or admin or nonadmin.

    Args:
      planId: The planId of the infrastructure.

    Returns:
      context: The k8s context.
    """
    tenant = None
    tenant_id = self.duplo.tenantid
    tenant_name = self.duplo.tenant
    identified = True if tenant_id or tenant_name else False
    admin = self.duplo.isadmin

    # don't even like try sometimes
    if not admin and not identified:
      raise DuploError("--tenant is required", 300)
    if not planId and not identified:
      raise DuploError("--plan or --tenant is required", 300)

    # and admin needs a plan and may have a identified a tenant
    # or maybe it's not an admin and a name was used to identify the tenant
    if (admin and not planId) or (not admin and not tenant_id):
      tenant_svc = self.duplo.load("tenant")
      tenant = tenant_svc.find(tenant_name)
      planId = tenant.get("PlanID")
      tenant_id = tenant.get("TenantId")

    # choose the correct path
    path = (f"v3/admin/plans/{planId}/k8sConfig"
            if admin
            else f"/v3/subscriptions/{tenant_id}/k8s/jitAccess")
    response = self.duplo.get(path)
    return response.json()

  def __k8s_exec_credential(self, ctx):
    cluster = {
      "server": ctx["ApiServer"],
      "config": None
    }

    # only add the ca if it is present, Azure won't have one
    if (ca := ctx.get("CertificateAuthorityDataBase64", None)):
      cluster["certificate-authority-data"] = ca

    t = jwt.decode(jwt=ctx["Token"],algorithms=["HS256"],options={"verify_signature": False})
    exp = datetime.fromtimestamp(t["exp"]).strftime('%Y-%m-%dT%H:%M:%S+00:00')

    return {
      "kind": "ExecCredential",
      "apiVersion": "client.authentication.k8s.io/v1beta1",
      "spec": {
        "cluster": cluster,
        "interactive": False
      },
      "status": {
        "token": ctx["Token"],
        "expirationTimestamp": exp
      }
    }

  def __cluster_config(self, ctx):
    """Build a kubeconfig cluster object"""
    cluster = {
      "server": ctx["ApiServer"]
    }
    if (ca := ctx.get("CertificateAuthorityDataBase64", None)):
      cluster["certificate-authority-data"] = ca
    else:
      cluster["insecure-skip-tls-verify"] = True
    return {
      "name": ctx["PlanID"],
      "cluster": cluster
    }

  def __user_config(self, ctx):
    """Build a kubeconfig user object"""
    cmd = self.duplo.build_command(*ctx["ARGS"])
    return {
      "name": ctx["Name"],
      "user": {
        "exec": {
          "apiVersion": "client.authentication.k8s.io/v1beta1",
          "installHint": INSTALL_HINT,
          "command": "duploctl",
          "args": cmd
        }
      }
    }

  def __context_config(self, ctx):
    """Build a kubeconfig context object"""
    return {
      "name": ctx["Name"],
      "context": {
        "cluster": ctx["PlanID"],
        "user": ctx["Name"],
        "namespace": ctx["DefaultNamespace"]
      }
    }

  def __add_to_kubeconfig(self, section, item, kubeconfig):
    """Add an item to a kubeconfig section if it is not already present"""
    exists = False
    for i in kubeconfig[section]:
      if i["name"] == item["name"]:
        exists = True
        i.update(item)
        break
    if not exists:
      kubeconfig[section].append(item)

  def __empty_kubeconfig(self):
    return {
      "apiVersion": "v1",
      "kind": "Config",
      "preferences": {},
      "clusters": [],
      "users": [],
      "contexts": []
    }



import pytest
import time
from duplocloud.errors import DuploError
from .conftest import get_test_data

@pytest.fixture(scope="class")
def hosts_resource(duplo, request):
    """Fixture to load the Hosts resource and define host name."""
    resource = duplo.load("hosts")
    resource.duplo.wait = True
    tenant = resource.tenant["AccountName"]
    request.cls.host_name = f"duploservices-{tenant}-duplohost"
    return resource

def execute_test(func, *args, **kwargs):
    """Helper function to execute a test and handle errors."""
    try:
        return func(*args, **kwargs)
    except DuploError as e:
        pytest.fail(f"Test failed: {e}")

class TestHosts:

    @pytest.mark.integration
    @pytest.mark.dependency(name="create_host", scope="session")
    @pytest.mark.order(5)
    def test_create_host(self, hosts_resource):
        body = get_test_data("hosts")
        response = execute_test(hosts_resource.create, body)
        assert "Successfully created host" in response["message"]
        time.sleep(60)

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_host"], scope="session")
    @pytest.mark.order(6)
    def test_find_host(self, hosts_resource):
        host = execute_test(hosts_resource.find, self.host_name)
        assert host["FriendlyName"] == self.host_name
        assert host["Status"] == "running"

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_host"], scope="session")
    @pytest.mark.order(7)
    def test_stop_host(self, hosts_resource):
        response = execute_test(hosts_resource.stop, self.host_name)
        assert "Successfully stopped host" in response["message"]
        host = execute_test(hosts_resource.find, self.host_name)
        assert host["Status"] == "stopped"

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_host"], scope="session")
    @pytest.mark.order(8)
    def test_start_host(self, hosts_resource):
        response = execute_test(hosts_resource.start, self.host_name)
        assert "Successfully started host" in response["message"]
        host = execute_test(hosts_resource.find, self.host_name)
        assert host["Status"] == "running"

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_host"], scope="session")
    @pytest.mark.order(9)
    def test_reboot_host(self, hosts_resource):
        response = execute_test(hosts_resource.reboot, self.host_name)
        assert "Successfully rebooted host" in response["message"]
        time.sleep(60)
        host = execute_test(hosts_resource.find, self.host_name)
        assert host["Status"] == "running"

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_host"], scope="session")
    @pytest.mark.order(997)
    def test_delete_host(self, hosts_resource):
        response = execute_test(hosts_resource.delete, self.host_name)
        assert "Successfully deleted host" in response["message"]



import configparser
import os
import pytest
from duplocloud.client import DuploClient
from duplocloud.errors import DuploError

class TestJIT:
  
  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_admin_aws_jit(self, duplo: DuploClient):
    duplo.isadmin = True
    r = duplo.load("jit")
    try:
      o = r.aws()
      assert o.get("SecretAccessKey", None) is not None
    except DuploError as e:
      pytest.fail(f"Failed to get admin credentials: {e}")
  
  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_user_aws_jit(self, duplo: DuploClient):
    duplo.isadmin = False
    r = duplo.load("jit")
    try:
      o = r.aws()
      assert o.get("SecretAccessKey", None) is not None
    except DuploError as e:
      pytest.fail(f"Failed to get user credentials: {e}")
  
  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_user_k8s_context(self, duplo: DuploClient):
    duplo.isadmin = False
    r = duplo.load("jit")
    try:
      o = r.k8s_context()
      assert o.get("Token", None) is not None
    except DuploError as e:
      pytest.fail(f"Failed to get user k8s credentials: {e}")
  
  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_admin_k8s_context(self, duplo: DuploClient):
    duplo.isadmin = True
    r = duplo.load("jit")
    try:
      o = r.k8s_context()
      assert o.get("Token", None) is not None
    except DuploError as e:
      pytest.fail(f"Failed to get admin k8s credentials: {e}")

  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_update_aws_config_nonexistent_dir(self, duplo: DuploClient, tmp_path):
    config_dir = tmp_path / "nonexistent" / "nested"
    config_file = config_dir / "config"
    os.environ["AWS_CONFIG_FILE"] = str(config_file)
    r = duplo.load("jit")
    profile_name = "test_profile"
    assert not config_dir.exists()
    result = r.update_aws_config(profile_name)
    assert "added" in result["message"].lower()
    assert config_dir.exists()
    assert config_file.exists()
    config = configparser.ConfigParser()
    config.read(config_file)
    profile_section = f"profile {profile_name}"
    assert config.has_section(profile_section)
    os.environ.pop("AWS_CONFIG_FILE", None)

  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_update_kubeconfig(self, duplo: DuploClient, tmp_path):
    kubeconfig_file = tmp_path / "config"
    os.environ["KUBECONFIG"] = str(kubeconfig_file)
    r = duplo.load("jit")
    def mock_k8s_context(planId=None):
      return {
        "Name": "test-cluster",
        "ApiServer": "https://test-cluster:6443",
        "Token": "test-token",
        "DefaultNamespace": "default"
      }
    r.k8s_context = mock_k8s_context
    result = r.update_kubeconfig(planId="test-plan")
    assert "successfully" in result["message"].lower()
    assert kubeconfig_file.exists()
    import yaml
    with open(kubeconfig_file) as f:
      config = yaml.safe_load(f)
    assert "clusters" in config
    assert "users" in config
    assert "contexts" in config
    assert config["current-context"] == "test-plan"
    os.environ.pop("KUBECONFIG", None)

  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_update_kubeconfig_nonexistent_dir(self, duplo: DuploClient, tmp_path):
    kube_dir = tmp_path / "nonexistent" / "nested" / ".kube"
    kubeconfig_file = kube_dir / "config"
    os.environ["KUBECONFIG"] = str(kubeconfig_file)
    r = duplo.load("jit")
    def mock_k8s_context(planId=None):
      return {
        "Name": "test-cluster",
        "ApiServer": "https://test-cluster:6443",
        "Token": "test-token",
        "DefaultNamespace": "default"
      }
    r.k8s_context = mock_k8s_context
    assert not kube_dir.exists()
    result = r.update_kubeconfig(planId="test-plan")
    assert "successfully" in result["message"].lower()
    assert kube_dir.exists()
    assert kubeconfig_file.exists()
    import yaml
    with open(kubeconfig_file) as f:
      config = yaml.safe_load(f)
    assert "clusters" in config
    assert "users" in config
    assert "contexts" in config
    os.environ.pop("KUBECONFIG", None)

  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_update_aws_config(self, duplo: DuploClient, tmp_path):
    config_file = tmp_path / "config"
    os.environ["AWS_CONFIG_FILE"] = str(config_file)
    r = duplo.load("jit")
    profile_name = "test_profile"
    result = r.update_aws_config(profile_name)
    assert "added" in result["message"].lower()
    config = configparser.ConfigParser()
    config.read(config_file)
    profile_section = f"profile {profile_name}"
    assert config.has_section(profile_section)
    assert config.get(profile_section, "region") == os.getenv("AWS_DEFAULT_REGION", "us-west-2")
    assert "duploctl jit aws" in config.get(profile_section, "credential_process")
    result = r.update_aws_config(profile_name)
    assert "updated" in result["message"].lower()
    os.environ.pop("AWS_CONFIG_FILE", None)



import pytest
from unittest.mock import ANY, MagicMock
from duplo_resource.ecs_service import DuploEcsService
from duplocloud.errors import DuploError
from tests.test_utils import execute_test, assert_response

@pytest.mark.unit
def test_update_image_with_container(mocker):
    mock_client = mocker.MagicMock()
    service = DuploEcsService(mock_client)
    # Mock task definition with multiple containers
    mock_task_def = {
        "ContainerDefinitions": [
            {"Name": "app", "Image": "old-image:1"},
            {"Name": "sidecar", "Image": "sidecar:1"}
        ]
    }
    # Mock service family
    mock_service_family = {
        "DuploEcsService": {
            "TaskDefinition": "old-arn"
        }
    }
    # Mock the required methods
    mocker.patch.object(service, 'prefixed_name', return_value="test-service")
    mocker.patch.object(service, 'find_def', return_value=mock_task_def)
    mocker.patch.object(service, 'update_taskdef', return_value={"arn": "new-arn"})
    mocker.patch.object(service, 'find_service_family', return_value=mock_service_family)
    mocker.patch.object(service, 'update_service')
    mocker.patch.object(service, 'wait')
    # Enable wait flag
    mock_client.wait = True
    # Test updating specific container
    result = execute_test(service.update_image, "test-service", container_image=[("sidecar", "new-image:2")])
    # Verify the container image was updated
    assert mock_task_def["ContainerDefinitions"][1]["Image"] == "new-image:2"
    assert mock_task_def["ContainerDefinitions"][0]["Image"] == "old-image:1"
    # Verify service was updated with new task definition
    service.update_service.assert_called_once_with(mock_service_family["DuploEcsService"])
    assert_response(result, "Updating a task definition and its corresponding service.")

@pytest.mark.unit
def test_update_image_without_container(mocker):
    mock_client = mocker.MagicMock()
    service = DuploEcsService(mock_client)
    # Mock task definition with single container
    mock_task_def = {
        "ContainerDefinitions": [
            {"Name": "app", "Image": "old-image:1"}
        ]
    }
    # Mock service family
    mock_service_family = {
        "DuploEcsService": {
            "TaskDefinition": "old-arn"
        }
    }
    # Mock the required methods
    mocker.patch.object(service, 'prefixed_name', return_value="test-service")
    mocker.patch.object(service, 'find_def', return_value=mock_task_def)
    mocker.patch.object(service, 'update_taskdef', return_value={"arn": "new-arn"})
    mocker.patch.object(service, 'find_service_family', return_value=mock_service_family)
    mocker.patch.object(service, 'update_service')
    mocker.patch.object(service, 'wait')
    # Enable wait flag
    mock_client.wait = True
    # Test updating without specifying container (should update first container)
    result = execute_test(service.update_image, "test-service", image="new-image:2")
    # Verify the first container image was updated
    assert mock_task_def["ContainerDefinitions"][0]["Image"] == "new-image:2"
    # Verify service was updated with new task definition
    service.update_service.assert_called_once_with(mock_service_family["DuploEcsService"])
    assert_response(result, "Updating a task definition and its corresponding service.")

@pytest.mark.unit
def test_update_image_no_service(mocker):
    mock_client = mocker.MagicMock()
    service = DuploEcsService(mock_client)
    # Mock task definition
    mock_task_def = {
        "ContainerDefinitions": [
            {"Name": "app", "Image": "old-image:1"}
        ]
    }
    # Mock the required methods
    mocker.patch.object(service, 'prefixed_name', return_value="test-service")
    mocker.patch.object(service, 'find_def', return_value=mock_task_def)
    mocker.patch.object(service, 'update_taskdef', return_value={"arn": "new-arn"})
    mocker.patch.object(service, 'find_service_family', side_effect=DuploError("Service not found"))
    # Test updating image when no service exists
    result = execute_test(service.update_image, "test-service", image="new-image:2")
    # Verify the image was updated in task definition
    assert mock_task_def["ContainerDefinitions"][0]["Image"] == "new-image:2"
    # Verify appropriate message is returned
    assert_response(result, "No Service Configured, only the definition is updated.")



import os
import pytest

from duplocloud.errors import DuploError
from duplocloud.client import DuploClient

# current working directory as variable
cwd = os.getcwd()
host = "http://example.duplocloud.net/nothing/?foo=bar"
cache_dir = f"{cwd}/.tmp/cache"

@pytest.mark.unit
def test_new_config():
  c = DuploClient(host=host)
  assert c.host == "https://example.duplocloud.net"

@pytest.mark.unit
def test_at_least_host(mocker):
  """No Host Gets Error"""

  # Patch in a mock of the client's config file so this test doesn't depend on a specific local setup.
  # Alternatively, we could set the config_file arg of DuploClient to a YAML in the tests/files directory. A mock
  # exercises the constructor's default (and requires no changes to the test this was added to fix).
  duplo_config_file = '''
---
current-context: this
contexts:
- name: this
  tenant: none
  token: none

  # Host not set to intentionally trigger the expected error.
  # host: localhost
'''
  mocker.patch('builtins.open', mocker.mock_open(read_data=duplo_config_file))

  duplo = DuploClient()
  with pytest.raises(DuploError) as e:
    duplo.token
    print(e)

@pytest.mark.unit
def test_cache_dir():
  c = DuploClient(
    host=host,
    cache_dir=cache_dir)
  assert c.cache_dir == cache_dir
  random_data = {"foo": "bar"}
  cf = f"{cache_dir}/test.json"
  c.set_cached_item("test", random_data)
  assert os.path.exists(cf), f"Cache file {cf} not found"
  # now check if we can get the data back
  assert c.get_cached_item("test") == random_data, "Cached data does not match"
  # delete the cache file
  os.remove(cf)
  os.rmdir(cache_dir)



from duplocloud.client import DuploClient
from duplocloud.errors import DuploError
from duplocloud.resource import DuploResource
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("plan")
class DuploPlan(DuploResource):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
  
  @Command()
  def list(self) -> list:
    """List of Plans
    
    Usage: CLI Usage
      ```bash
      duploctl plan list
      ```
    
    Returns:
      list: List of Plans
    """
    response = self.duplo.get("adminproxy/GetPlans")
    return response.json()
  
  @Command()
  def find(self, 
           name: args.NAME) -> dict:
    """Find a Plan by name

    Usage: CLI Usage
      ```bash
      duploctl plan find <name>
      ```
    
    Args:
      name: Plan Name

    Returns:
      dict: Plan Details
    """
    try:
      return [s for s in self.list() if self.name_from_body(s) == name][0]
    except IndexError:
      raise DuploError(f"Plan '{name}' not found", 404)

  def name_from_body(self, body):
    return body["Name"]



import pytest
import time
from duplocloud.errors import DuploError

@pytest.fixture(scope="class")
def aws_secret_resource(duplo):
    """Fixture to load the AWS Secret resource and define the secret name."""
    resource = duplo.load("aws_secret")
    resource.duplo.wait = True
    tenant = resource.tenant["AccountName"]
    secret_name = f"duploservices-{tenant}-secret"
    return resource, secret_name

def execute_test(func, *args, **kwargs):
    """Helper function to execute a test and handle errors."""
    try:
        return func(*args, **kwargs)
    except DuploError as e:
        pytest.fail(f"Test failed: {e}")

class TestAwsSecret:

    @pytest.mark.integration
    @pytest.mark.dependency(name="create_secret", scope="session")
    @pytest.mark.order(1)
    def test_create_secret(self, aws_secret_resource):
        """Test creating an AWS secret."""
        r, secret_name = aws_secret_resource
        body = {"Name": secret_name, "SecretString": '{"foo": "bar"}'}
        execute_test(r.create, name=secret_name, body=body)
        time.sleep(10)

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_secret"], scope="session")
    @pytest.mark.order(2)
    def test_find_secret(self, aws_secret_resource):
        """Test finding the created AWS secret."""
        r, secret_name = aws_secret_resource
        secret = execute_test(r.find, secret_name, show_sensitive=True)
        assert secret["Name"] == secret_name
        assert secret["SecretString"] == '{"foo": "bar"}'

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_secret"], scope="session")
    @pytest.mark.order(3)
    def test_update_secret(self, aws_secret_resource):
        """Test updating an AWS secret and verifying the update."""
        r, secret_name = aws_secret_resource
        new_value = '{"foo": "baz"}'
        execute_test(r.update, name=secret_name, value=new_value)
        # Verify the updated value
        updated_secret = execute_test(r.find, secret_name, show_sensitive=True)
        assert "SecretString" in updated_secret, "SecretString key missing in response"
        assert updated_secret["SecretString"] == new_value

    @pytest.mark.integration
    @pytest.mark.dependency(depends=["create_secret"], scope="session")
    @pytest.mark.order(4)
    def test_delete_secret(self, aws_secret_resource):
        """Test deleting an AWS secret."""
        r, secret_name = aws_secret_resource
        response = execute_test(r.delete, secret_name)
        assert response.get("message") == f"Successfully deleted secret '{secret_name}'"
